#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
import subprocess
from typing import Optional
from splunk_add_on_ucc_modinput_test.common import utils
import json


def base64encode(
    text_file: Path,
    string: str,
) -> str:
    if string is None:
        string = text_file.read_text()
    return utils.Base64.encode(string=string)


def base64decode(base64_string: str) -> str:
    return utils.Base64.decode(base64_string=base64_string)


def get_rest_root(*, openapi: Path) -> str:
    with openapi.open() as f:
        data = json.load(f)
    paths = list(data.get("paths", {}).keys())
    common_prefix = utils.find_common_prefix(paths)
    if (
        common_prefix is None
        or common_prefix == ""
        or common_prefix[0] != "/"
        or common_prefix[-1] != "_"
    ):
        raise ValueError("Invalid common prefix")
        return ""
    return common_prefix[1:-1]


def is_docker_running() -> Optional[str]:
    """
    Checks if Docker is running on the system.

    Returns:
        Optional[str]:
            - None if Docker is running without issues.
            - A string describing if Docker is not running or not installed.
    """
    try:
        subprocess.check_output(
            ["docker", "--version"], stderr=subprocess.STDOUT
        )
    except Exception:
        return "Docker not installed"
    try:
        subprocess.check_output(["docker", "info"], stderr=subprocess.STDOUT)
        return None
    except Exception:
        return "Docker seems to be installed but not started"



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from splunk_add_on_ucc_modinput_test.typing import (
        ExecutableKeyType,
        ProbeFnType,
        ForgeFnType,
        TestFnType,
    )
import inspect
from typing import Any


class ExecutableBase:
    def __init__(
        self,
        function: ProbeFnType | ForgeFnType | TestFnType,
    ) -> None:
        assert callable(function)
        self._function = function
        self._inspect()

    @property
    def source_file(self) -> str:
        return self._fn_source_file

    @property
    def key(self) -> ExecutableKeyType:
        return (self._fn_source_file, self.fn_full_name)

    @property
    def original_key(self) -> ExecutableKeyType:
        return (self._fn_source_file, self.fn_original_full_name)

    @property
    def original_name(self) -> str:
        if self._original_name == "__call__":
            return (
                self._fn_bound_class.lower()
                if self._fn_bound_class
                else "__call__"
            )
        return self._original_name

    @property
    def fn_full_name(self) -> str:
        if self._fn_bound_class:
            return f"{self._fn_bound_class}::{self._fn_name}"
        else:
            return self._fn_name

    @property
    def fn_original_full_name(self) -> str:
        if self._fn_bound_class:
            return f"{self._fn_bound_class}::{self._original_name}"
        else:
            return self._original_name

    def _inspect(self) -> None:
        self._fn_bound_class: str | None = None
        if inspect.ismethod(self._function):
            self._fn_bound_class = self._function.__self__.__class__.__name__
            self._fn_name = self._function.__name__
            self._fn_source_file = inspect.getfile(self._function)
        elif inspect.isfunction(self._function):
            self._fn_bound_class = None
            res = repr(self._function).split(" ")[1].split(".")
            if len(res) > 1:
                self._fn_bound_class = res[0]
                self._fn_name = res[1]
            else:
                self._fn_name = res[0]
            self._fn_source_file = inspect.getfile(self._function)
        else:
            self._fn_name = "__call__"
            self._fn_bound_class = self._function.__class__.__name__
            self._fn_source_file = inspect.getfile(self._function.__class__)

        self._original_name = self._fn_name
        self._is_generatorfunction = inspect.isgeneratorfunction(
            self._function
        )

        sig = inspect.signature(self._function)
        self._required_args = list(sig.parameters.keys())

    @property
    def required_args_names(self) -> tuple[str, ...]:
        return tuple(self._required_args)

    def filter_requied_kwargs(self, kwargs: dict[str, Any]) -> dict[str, Any]:
        return {k: v for k, v in kwargs.items() if k in self._required_args}



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# mypy: disable-error-code="attr-defined"

import os
import time
import datetime
from functools import lru_cache
import pytz  # type: ignore
import base64
from pathlib import Path
from typing import Callable, List, Optional
import hashlib
import logging

logger = logging.getLogger("ucc-modinput-test")


class SplunkClientConfigurationException(Exception):
    pass


def get_from_environment_variable(
    environment_variable: str,
    *,
    default_value: Optional[str] = None,
    is_optional: bool = False,
    string_function: Optional[Callable[[str], str]] = None,
) -> Optional[str]:
    def use_string_function_if_needed(
        *, variable: str, function: Optional[Callable[[str], str]] = None
    ) -> str:
        return variable if function is None else function(variable)

    if environment_variable in os.environ:
        return use_string_function_if_needed(
            variable=os.environ[environment_variable],
            function=string_function,
        )
    elif default_value is not None:
        return use_string_function_if_needed(
            variable=default_value,
            function=string_function,
        )
    elif is_optional:
        return None
    else:
        logger.critical(40 * "*")
        logger.critical(f"{environment_variable} environment variable not set")
        logger.critical("run below in terminal:")
        logger.critical(f"export {environment_variable}=[your value]")
        logger.critical(40 * "*")

        error = f"Mandatory environment variable {environment_variable} is \
            missing and does not have a default value specified."
        raise SplunkClientConfigurationException(error)


class Base64:
    @staticmethod
    def _remove_ending_chars(string: str) -> str:
        if len(string) == 0:
            return string
        CHARS_TO_BE_REMOVED = ["\n"]
        for i in range(len(string), -1, -1):
            if string[i - 1] not in CHARS_TO_BE_REMOVED:
                break
        return string[:i]

    @staticmethod
    def encode(string: str) -> str:
        _bytes = Base64._remove_ending_chars(string=string).encode("utf-8")
        base64_encoded = base64.b64encode(_bytes)
        base64_string = base64_encoded.decode("utf-8")
        return base64_string

    @staticmethod
    def decode(base64_string: str) -> str:
        base64_bytes = base64_string.encode("utf-8")
        decoded_bytes = base64.b64decode(base64_bytes)
        string = decoded_bytes.decode("utf-8")
        return Base64._remove_ending_chars(string=string)


def get_epoch_timestamp() -> float:
    return time.time()


@lru_cache(maxsize=32)
def convert_to_utc(
    epoch_timestamp: float, format: str = "%Y%m%d%H%M%S"
) -> str:
    return datetime.datetime.fromtimestamp(
        epoch_timestamp, pytz.timezone("UTC")
    ).strftime(format)


class Common:
    __instance = None

    def __new__(cls, *args, **kwargs):  # type: ignore
        if not Common.__instance:
            Common.__instance = object.__new__(cls)
            Common.__instance._start_timestamp = get_epoch_timestamp()
            logger.info(
                f"Test timestamp set to: \
                    {convert_to_utc(Common.__instance._start_timestamp)}"
            )
        return Common.__instance

    def __init__(self) -> None:
        pass

    @property
    def start_timestamp(self) -> float:
        return self._start_timestamp

    @start_timestamp.setter
    def start_timestamp(self, value: float) -> None:
        self._start_timestamp = value

    @property
    def sufix(self) -> str:
        return f"mit_{convert_to_utc(self.start_timestamp)}"
        # MIT from "MODULARINPUT TEST"


def find_common_prefix(strings: List[str]) -> Optional[str]:
    if strings is None or len(strings) == 0 or len(strings) == 1:
        return None
    elif len(strings) == 2:
        s0len = len(strings[0])
        s1len = len(strings[1])
        min_len = s0len if s0len < s1len else s1len
        for i in range(min_len):
            if strings[0][i] != strings[1][i]:
                return strings[0][:i]
        return strings[0][:min_len]
    else:
        common = find_common_prefix(strings[:2])
        if common is None or common == "":
            return common
        return find_common_prefix([common] + strings[2:])


def md5(*, file_path: Path, chunk_size: int = 8192) -> str:
    hash_md5 = hashlib.md5()
    with file_path.open("rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunklib import client
from threading import Lock
from typing import Any, Union, List
import logging

logger = logging.getLogger("ucc-modinput-test")


class SplunkServiceProxy:
    def __init__(
        self, host: str, port: Union[int, str], username: str, password: str
    ) -> None:
        self._lock = Lock()
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self.__connect()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._service, name)

    def __connect(self) -> None:
        with self._lock:
            self._service = client.connect(
                host=self._host,
                port=self._port,
                username=self._username,
                password=self._password,
            )


class SplunkServicePool:
    def __init__(
        self,
        host: str,
        port: Union[int, str],
        username: str,
        password: str,
        *,
        pool_initial_size: int = 3,
        pool_size_inc: int = 2,
    ) -> None:
        logger.debug(f"SplunkServicePool init {username}@{host}:{port}")

        self._lock = Lock()
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self._pool_initial_size = pool_initial_size
        self._pool_size_inc = pool_size_inc
        self._pool: List[SplunkServiceProxy] = []
        self.__increase_pool(self._pool_initial_size)

    def __increase_pool(self, increment_size: int) -> None:
        with self._lock:
            for _ in range(increment_size):
                self._pool.append(
                    SplunkServiceProxy(
                        self._host, self._port, self._username, self._password
                    )
                )
        logger.debug(f"SplunkServicePool has been sized to {len(self._pool)}")

    def __getattr__(self, name: str) -> Any:
        while True:
            for svc in self._pool:
                locked = svc._lock.acquire(False)
                if locked:
                    try:
                        return getattr(svc, name)
                    finally:
                        svc._lock.release()

            self.__increase_pool(self._pool_size_inc)
        return None



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional
from pytest import Config
from splunk_add_on_ucc_modinput_test.functional.constants import (
    Executor,
    ForgeProbe,
    TasksWait,
)


class PytestConfigAdapter:
    def __init__(self, pytest_config: Optional[Config] = None) -> None:
        self._pytest_config = pytest_config

    def link_pytest_config(
        self, pytest_config: Optional[Config] = None
    ) -> None:
        self._pytest_config = pytest_config

    @property
    def pytest_config(self) -> Optional[Config]:
        return self._pytest_config

    @property
    def do_not_fail_with_teardown(self) -> bool:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("do_not_fail_with_teardown")
        return False

    @property
    def sequential_execution(self) -> bool:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("sequential_execution")
        return False

    @property
    def number_of_threads(self) -> int:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("number_of_threads")
        return Executor.DEFAULT_THREAD_NUMBER.value

    @property
    def probe_invoke_interval(self) -> int:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("probe_invoke_interval")
        return ForgeProbe.DEFAULT_INTERVAL.value

    @property
    def probe_wait_timeout(self) -> int:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("probe_wait_timeout")
        return ForgeProbe.DEFAILT_WAIT_TIME.value

    @property
    def bootstrap_wait_timeout(self) -> int:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("bootstrap_wait_timeout")
        return TasksWait.DEFAULT_BOOTSTRAP_TIMEOUT.value

    @property
    def attached_tasks_wait_timeout(self) -> int:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("attached_tasks_wait_timeout")
        return TasksWait.DEFAULT_ATTACHED_TIMEOUT.value

    @property
    def completion_check_frequency(self) -> int:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("completion_check_frequency")
        return TasksWait.DEFAULT_CHECK_FREQUENCY.value

    @property
    def do_not_delete_at_teardown(self) -> bool:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("do_not_delete_at_teardown")
        return False

    @property
    def collectonly(self) -> bool:
        if self._pytest_config is not None:
            return self._pytest_config.getvalue("collectonly")
        return False



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunk_add_on_ucc_modinput_test.typing import (
    ExecutableKeyType,
)


from typing import Any, Callable, Generator, List, Dict, Optional, Tuple
from splunk_add_on_ucc_modinput_test.functional.entities.forge import (
    FrameworkForge,
)
from splunk_add_on_ucc_modinput_test.functional.entities.test import (
    FrameworkTest,
)
from splunk_add_on_ucc_modinput_test.functional.entities.task import (
    FrameworkTask,
    TaskSetListType,
)
from splunk_add_on_ucc_modinput_test.functional import logger


class TestCollection(Dict[ExecutableKeyType, FrameworkTest]):
    @property
    def is_empty(self) -> bool:
        return not bool(self)

    def add(self, item: FrameworkTest) -> None:
        assert isinstance(item, FrameworkTest)
        if item.key not in self:
            self[item.key] = item

    def lookup_by_function(
        self, fn: Callable[..., Any]
    ) -> Optional[FrameworkTest]:
        test = FrameworkTest(fn)
        return self.get(test.key)

    def lookup_by_original_function(
        self, fn: Callable[..., Any]
    ) -> List[ExecutableKeyType]:
        found_tests_keys = set()
        lookup_test = FrameworkTest(fn)
        for key, test in self.items():
            if lookup_test.key == test.original_key:
                logger.debug(
                    f"lookup_by_original_function found key: \
                        {lookup_test.key} -> {test.key}|{test.original_key}"
                )
                found_tests_keys.add(key)

        return list(found_tests_keys)


class ForgeCollection(Dict[ExecutableKeyType, FrameworkForge]):
    @property
    def is_empty(self) -> bool:
        return not bool(self)

    def add(self, item: FrameworkForge) -> None:
        if item.key not in self:
            self[item.key] = item


class TaskCollection:
    def __init__(self) -> None:
        self._tasks: Dict[ExecutableKeyType, TaskSetListType] = {}

    @property
    def is_empty(self) -> bool:
        return not bool(self._tasks)

    def remove_test_tasks(
        self, test_key: ExecutableKeyType
    ) -> Optional[TaskSetListType]:
        return self._tasks.pop(test_key, None)

    def add(self, tasks: List[FrameworkTask]) -> None:
        if not tasks:
            return
        test_key = tasks[0].test_key
        if test_key not in self._tasks:
            self._tasks[test_key] = []
        self._tasks[test_key].append(tasks)

    def get_tasks_by_type(
        self, test_key: ExecutableKeyType
    ) -> Tuple[TaskSetListType, TaskSetListType]:
        all_tasks = self.get_tasks(test_key)
        inplace_tasks: TaskSetListType = []
        bootstrap_tasks: TaskSetListType = []
        for step_tasks in all_tasks:
            if step_tasks is not None:
                if step_tasks[0].is_bootstrap:
                    bootstrap_tasks.append(step_tasks)
                else:
                    inplace_tasks.append(step_tasks)
        return inplace_tasks, bootstrap_tasks

    def get_bootstrap_tasks(
        self, test_key: ExecutableKeyType
    ) -> TaskSetListType:
        all_tasks = self.get_tasks(test_key)
        bootstrap_tasks: TaskSetListType = []
        for t in all_tasks:
            if t is not None and t[0].is_bootstrap:
                bootstrap_tasks.append(t)
        logger.debug(f"get_bootstrap_tasks for {test_key}: {bootstrap_tasks}")
        return bootstrap_tasks

    def get_inplace_tasks(
        self, test_key: ExecutableKeyType
    ) -> TaskSetListType:
        all_tasks = self.get_tasks(test_key)
        inplace_tasks: TaskSetListType = []
        for t in all_tasks:
            if t is not None and not t[0].is_bootstrap:
                inplace_tasks.append(t)
        logger.debug(f"get_inplace_tasks for {test_key}: {inplace_tasks}")
        return inplace_tasks

    def get_tasks(self, test_key: ExecutableKeyType) -> TaskSetListType:
        tasks = self._tasks.get(test_key, [])
        logger.debug(f"get_tasks for {test_key}: {tasks}")
        return tasks

    def enumerate_tasks(
        self, test_key: Tuple[str, ...]
    ) -> Generator[Tuple[int, int, FrameworkTask], None, None]:
        test_tasks = self.get_tasks(test_key)
        for i, parralel_tasks in enumerate(test_tasks):
            if parralel_tasks is not None:
                for j, task in enumerate(parralel_tasks):
                    yield i, j, task

    def enumerate_bootstrap_tasks(
        self, test_key: ExecutableKeyType
    ) -> Generator[Tuple[int, int, FrameworkTask], None, None]:
        bootstrap_tasks = self.get_bootstrap_tasks(test_key)
        for i, parralel_tasks in enumerate(bootstrap_tasks):
            if parralel_tasks is not None:
                for j, task in enumerate(parralel_tasks):
                    yield i, j, task

    def enumerate_inplace_tasks(
        self, test_key: ExecutableKeyType
    ) -> Generator[Tuple[int, int, FrameworkTask], None, None]:
        inplace_tasks = self.get_inplace_tasks(test_key)
        for i, parralel_tasks in enumerate(inplace_tasks):
            if parralel_tasks is not None:
                for j, task in enumerate(parralel_tasks):
                    yield i, j, task

    def bootstrap_tasks_by_state(
        self, test_key: ExecutableKeyType
    ) -> Tuple[List[FrameworkTask], List[FrameworkTask]]:
        done, pending = [], []
        for _, _, task in self.enumerate_bootstrap_tasks(test_key):
            if task.is_executed:
                done.append(task)
            else:
                pending.append(task)
        return done, pending



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from splunk_add_on_ucc_modinput_test.typing import (
    ArtifactsType,
    ExecutableKeyType,
)

import threading
from queue import Queue
import traceback
import time
from dataclasses import dataclass
from typing import Any, Callable, TYPE_CHECKING
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.entities.task import (
    FrameworkTask,
    TaskSetListType,
)
from splunk_add_on_ucc_modinput_test.functional.exceptions import (
    SplTaFwkWaitForDependenciesTimeout,
)

if TYPE_CHECKING:
    from splunk_add_on_ucc_modinput_test.functional.manager import (
        TestDependencyManager,
    )


def log_exceptions_traceback(fn: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(*args: Any, **kwargs: Any) -> Any | None:
        try:
            logger.debug(f"{fn.__name__} executor is about to start")
            return fn(*args, **kwargs)
        except Exception as e:
            logger.error(f"{fn.__name__} executor has failed with error: {e}")
            logger.debug(traceback.format_exc())
        finally:
            logger.debug(f"{fn.__name__} executor has stopped")
        return None

    return wrapper


class TaskGroupProcessor:
    @dataclass
    class Job:
        test_index: int
        task_index: int
        task: FrameworkTask

        @property
        def id(self) -> tuple[int, int]:
            return self.test_index, self.task_index

    def __init__(
        self,
        group: TaskSetListType,
        global_builtin_args_factory: Callable[
            [ExecutableKeyType], ArtifactsType
        ],
    ) -> None:
        self._global_builtin_args_factory = global_builtin_args_factory
        self._task_group: TaskSetListType = group
        self._jobs: list[TaskGroupProcessor.Job] = []
        self._matched_tasks: dict[tuple[int, int], tuple[int, int]] = {}
        self._result_collector: list[list[object | None] | None] = [
            None
        ] * len(self._task_group)
        self._done: dict[tuple[int, int], bool] = {}

        self._build_task_list()

    def _build_task_list(self) -> None:
        logger.debug(f"\npush task group {self._task_group}")

        for test_index, test_tasks in enumerate(self._task_group):
            logger.debug(
                f"test_index={test_index} push parallel_task {test_tasks}"
            )
            if test_tasks is not None:
                self._result_collector[test_index] = [None] * len(test_tasks)
                self._jobs += self._process_test_tasks(test_index, test_tasks)

        for job in self._jobs:
            self._done[job.id] = False

    @property
    def jobs(self) -> list[Job]:
        return self._jobs

    @property
    def all_tasks_done(self) -> bool:
        return all(self._done.values())

    def _process_test_tasks(
        self, test_index: int, test_tasks: list[FrameworkTask]
    ) -> list[Job]:
        processed_tasks = []
        for task_index, task in enumerate(test_tasks):
            try:
                task.prepare_forge_call_args(
                    self._global_builtin_args_factory(task.test_key)
                )
            except Exception as e:
                task.mark_as_failed(e, "Failed to prepare forge call args")
            else:
                if not self._try_skip_task(test_index, task_index):
                    job = self.Job(test_index, task_index, task)
                    processed_tasks.append(job)

        return processed_tasks

    def _get_task(self, test_index: int, task_index: int) -> FrameworkTask:
        test_tasks = self._task_group[test_index]
        assert test_tasks is not None, "test_tasks must be valid"
        return test_tasks[task_index]

    def _try_skip_task(
        self, test_index: int, task_index: int
    ) -> tuple[int, int] | None:
        task = self._get_task(test_index, task_index)
        same_task = self._find_same_task(task)
        logger.debug(f"{task.forge_key} SAME TASK IS  {same_task}")
        if same_task:
            logger.debug(
                f"task_index={task_index} skip task {task.forge_key} execution"
            )
            self._matched_tasks[(test_index, task_index)] = same_task
        return same_task

    def _find_same_task(self, task: FrameworkTask) -> tuple[int, int] | None:
        for i, test_tasks in enumerate(self._task_group):
            if test_tasks is None:
                continue
            for j, t in enumerate(test_tasks):
                if task is t:
                    return None
                if task.same_tasks(t):
                    return i, j
        return None

    def _update_test_artifacts(self, test_index: int) -> None:
        test_results = self._result_collector[test_index]
        if test_results:
            for task_index, test_result in enumerate(test_results):
                task = self._get_task(test_index, task_index)
                logger.debug(
                    f"TEST RESULT {test_result} is dict - {isinstance(test_result, dict)}, test {task.test_key}, dep: {task.forge_key}, result {task.result}"
                )
                test_result = task.make_kwarg(test_result)
                task.update_test_artifacts(test_result)
                logger.debug(
                    f"ARTIFACTS UPDATED dep: {task.forge_key}, test {task.test_key}, artifacts: {task._test.artifacts}"
                )

    def _copy_result_to_matching_tasks(
        self, src_test_i: int, src_task_j: int
    ) -> None:
        src_task = self._get_task(src_test_i, src_task_j)
        for dst, src in self._matched_tasks.items():
            if src == (src_test_i, src_task_j):
                dst_test_i, dst_task_j = dst

                row = self._result_collector[dst_test_i]
                if row is not None:
                    row[dst_task_j] = src_task._result

                    dst_task = self._get_task(dst_test_i, dst_task_j)
                    if src_task._exec_id is not None:
                        dst_task.reuse_forge_execution(
                            src_task._exec_id,
                            src_task._result,
                            src_task._setup_errors,
                        )
                        dst_task.mark_as_executed()
                        self._update_test_artifacts(dst_test_i)

    def process_response(self, job: Job) -> None:
        finished_task = self._get_task(job.test_index, job.task_index)
        assert finished_task is job.task, "Task is not the same!"
        logger.debug(
            f"manager got finished task {job.id}, TEST KEY: {finished_task.test_key}, DEP KEY: {finished_task.forge_key}, DEP RES: {finished_task.result}, RES {job.task.result}"
        )

        row = self._result_collector[job.test_index]
        if row is not None:
            row[job.task_index] = job.task.result
            self._update_test_artifacts(job.test_index)
            self._copy_result_to_matching_tasks(job.test_index, job.task_index)
            self._done[job.id] = True
            logger.debug(
                f"manager is waiting for tasks {self._done}, results: {self._result_collector}"
            )


class FrmwkExecutorBase:
    def __init__(self, manager: TestDependencyManager) -> None:
        self._manager = manager

    def shutdown(self) -> None:
        pass

    def wait(self, is_bootstrap: bool) -> None:
        pass

    def global_builtin_args_factory(
        self, test_key: ExecutableKeyType
    ) -> ArtifactsType:
        return self._manager.get_global_builtin_args(test_key)

    def _execute_request(
        self, job: TaskGroupProcessor.Job, worker_id: int = 0
    ) -> None:
        try:
            task_info = f"{job.id}, task: {id(job.task)} - {job.task}, dep: {id(job.task._forge)} - {job.task._forge} - {job.task.forge_key}, call_args: {job.task._forge_kwargs}"
            logger.debug(f"worker {worker_id}, task started {task_info}")
            job.task.execute()
        except Exception as e:
            logger.debug(
                f"worker {worker_id}, task failed {task_info} with error {e}\n{traceback.format_exc()}"
            )
        else:
            logger.debug(
                f"worker {worker_id}, task finished {task_info} with result: {job.task.result}"
            )


class FrmwkSequentialExecutor(FrmwkExecutorBase):
    def start(self, tasks: list[TaskSetListType]) -> None:
        logger.debug(f"sequential executor has started with tasks {tasks}")
        for task_group in tasks:
            proc = TaskGroupProcessor(
                task_group, self.global_builtin_args_factory
            )
            for job in proc.jobs:
                self._execute_request(job)
                proc.process_response(job)

        logger.debug("sequential executor has finished")


class FrmwkParallelExecutor(FrmwkExecutorBase):
    def __init__(
        self, manager: TestDependencyManager, worker_count: int = 10
    ) -> None:
        super().__init__(manager)
        self.worker_count = worker_count
        self.task_queue: Queue[TaskGroupProcessor.Job | None] = Queue()
        self._manager_queue: Queue[
            list[TaskSetListType] | TaskGroupProcessor.Job | None
        ] = Queue()
        self._is_free: threading.Event = threading.Event()
        self.deploy()

    def deploy(self) -> None:
        self.threads = [
            threading.Thread(target=self.worker_thread, args=(index,))
            for index in range(self.worker_count)
        ]

        for thread in self.threads:
            thread.start()

        self._manager_thread = threading.Thread(target=self.manager_thread)
        self._manager_thread.start()

        logger.debug("deployed")
        self._is_free.set()
        logger.debug("FrmwkParallelExecutor is set free")

    def wait(self, is_bootstrap: bool = True) -> None:
        if is_bootstrap:
            wait_timeout = self._manager.bootstrap_wait_timeout
            task_type = "bootstrap"
        else:
            wait_timeout = self._manager.attached_tasks_wait_timeout
            task_type = "attached"

        expiration = time.time() + wait_timeout
        while not self._is_free.wait(self._manager.completion_check_frequency):
            if time.time() > expiration:
                msg = f"Waiting for executor to process all {task_type} tasks exceeded timeout {wait_timeout} seconds."
                logger.error(msg)
                raise SplTaFwkWaitForDependenciesTimeout(msg)
            logger.debug(
                f"Still waiting for executor to process all {task_type} tasks"
            )

    def start(self, tasks: list[TaskSetListType]) -> None:
        logger.debug("FrmwkParallelExecutor::start - wait for previoues tasks")
        self.wait()
        logger.debug(
            "FrmwkParallelExecutor::start - executer is about to set busy"
        )
        self._is_free.clear()
        logger.debug("FrmwkParallelExecutor is set busy")
        logger.debug("FrmwkParallelExecutor::start - sending tasks to manager")
        self._manager_queue.put(tasks)
        logger.debug("started")

    def shutdown(self) -> None:
        logger.debug("Sending shutdown command to executor...")
        self._manager_queue.put(None)
        logger.debug("Waiting for executor to shutdown...")
        self._manager_thread.join()
        logger.info("Executor has shutdown.")

    def _receive_tasks(
        self,
    ) -> tuple[bool, list[TaskSetListType] | TaskGroupProcessor.Job | None]:
        logger.debug(
            "FrmwkParallelExecutor::_receive_tasks - set waiting for tasks"
        )
        tasks = self._manager_queue.get()
        logger.debug(
            f"FrmwkParallelExecutor::_receive_tasks - got tasks: {tasks}"
        )
        self._manager_queue.task_done()
        logger.debug("FrmwkParallelExecutor::_receive_tasks - queue task done")
        if tasks is None:
            return True, []
        return False, tasks

    def _collect_results(self, proc: TaskGroupProcessor) -> bool:
        while not proc.all_tasks_done:
            job = self._manager_queue.get()
            logger.debug(f"manager got finished task {job}")
            if job is None:
                logger.debug("manager is interrupted")
                return True

            if isinstance(job, TaskGroupProcessor.Job):
                proc.process_response(job)
                self._manager_queue.task_done()

        return False

    @log_exceptions_traceback
    def manager_thread(self) -> None:
        logger.debug("manager has started")
        interrupted = False
        while not interrupted:
            interrupted, tasks = self._receive_tasks()
            logger.debug(f"manager got tasks {tasks}")
            if isinstance(tasks, list):
                for task_group in tasks:
                    proc = TaskGroupProcessor(
                        task_group, self.global_builtin_args_factory
                    )
                    for job in proc.jobs:
                        self.task_queue.put(job)

                    interrupted = self._collect_results(proc)
                    if interrupted:
                        break

            self._is_free.set()
            logger.debug("FrmwkParallelExecutor is set free")

        for _ in self.threads:
            self.task_queue.put(None)

        for thread in self.threads:
            thread.join()

        logger.debug("manager has finished")

    @log_exceptions_traceback
    def worker_thread(self, wid: int) -> None:
        while True:
            job = self.task_queue.get()
            logger.debug(f"worker {wid} task recieved {job}")
            if job is None:
                break

            self._execute_request(job, wid)
            self.task_queue.task_done()
            self._manager_queue.put(job)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
from typing import Tuple, List, Dict
from jinja2 import Environment, Template, select_autoescape, FileSystemLoader
from splunk_add_on_ucc_modinput_test import resources
from importlib_resources import files


CLIENT_FILE_NAME = "_managed_client.py"


def get_splunk_client_path(ucc_modinput_functional: Path) -> Path:
    return ucc_modinput_functional / "splunk" / "client" / CLIENT_FILE_NAME


def load_readme_examples(swagger_client_readme_md: Path) -> List[str]:
    split_marker = "# create an instance of the API class"
    end_marker = "## Documentation for API Endpoints"

    with open(swagger_client_readme_md) as file:
        content = file.read()

    from_pos = content.find(split_marker)
    to_pos = content.find(end_marker)
    if from_pos < 0:
        raise ValueError(
            f"Could not find start marker '{split_marker}' in README.md"
        )
    if to_pos < 0:
        raise ValueError(
            f"Could not find end marker '{end_marker}' in README.md"
        )
    if to_pos < from_pos:
        raise ValueError(
            f"End marker '{end_marker}' is before start marker \
                '{split_marker}' in README.md"
        )

    splited_content = content[from_pos:to_pos].split(split_marker)[
        1:
    ]  # eliminates everything before the first split_marker
    return [x.strip() for x in splited_content]


def get_jinja_env() -> Environment:
    templates_search_path = files(resources).joinpath("templates")
    return Environment(
        loader=FileSystemLoader(templates_search_path),
        autoescape=select_autoescape(),
    )


def remove_prefix(*, api_name: str, ta_api_prefix: str) -> str:
    if api_name.lower().startswith(ta_api_prefix.lower()):
        skip = len(ta_api_prefix)
        api_name_no_prefix = api_name[skip:]
        while api_name_no_prefix and not api_name_no_prefix[0].isalpha():
            api_name_no_prefix = api_name_no_prefix[1:]
        return api_name_no_prefix
    return api_name


def make_method_name(
    *, api_name_no_prefix: str, args_specs: Dict[str, List[str]]
) -> Tuple[str, str]:
    if api_name_no_prefix.endswith("_name_get"):
        if "name" in args_specs:
            return ("get_" + api_name_no_prefix[:-9], "get")
        else:
            return ("get_" + api_name_no_prefix[:-4], "get")

    if api_name_no_prefix.endswith("_get"):
        for check in ["proxy", "settings", "logging"]:
            if check in api_name_no_prefix:
                return ("get_" + api_name_no_prefix[:-4], "get")

        if "name" not in args_specs:
            return ("get_" + api_name_no_prefix[:-4] + "_list", "list")

    if api_name_no_prefix.endswith("_name_post"):
        if "name" in args_specs:
            return ("update_" + api_name_no_prefix[:-10], "create")
        else:
            return ("create_" + api_name_no_prefix[:-5], "create")

    if api_name_no_prefix.endswith("_post"):
        for check in ["proxy", "settings", "logging"]:
            if check in api_name_no_prefix:
                return ("update_" + api_name_no_prefix[:-5], "update")

        return ("create_" + api_name_no_prefix[:-5], "create")

    if api_name_no_prefix.endswith("_name_delete"):
        if "name" in args_specs:
            return ("delete_" + api_name_no_prefix[:-12], "delete")
        else:
            return ("delete_" + api_name_no_prefix[:-7], "delete")

    return api_name_no_prefix, "unknown"


# def _parse_arg_descriptor(self, arg_name: str, right_part: str) -> None:
def parse_arg_descriptor(right_part: str) -> List[str]:
    descriptors = right_part.strip().split("#")
    spec = []
    if len(descriptors) > 1:
        spec_str = descriptors[1].strip()
        for item in spec_str.split("|"):
            spec.append(item.strip())
    # self.args_specs[arg_name] = spec
    return spec


def parse_args(
    *, call_args: List[str], args_specs: Dict[str, List[str]]
) -> Tuple[str, str]:
    api_args = []
    method_args = []
    method_kwargs = []
    for arg in call_args:
        arg = arg.strip()
        kv = arg.split("=")
        if len(kv) > 1:
            arg_name = kv[0].strip()
            api_arg_str = f"{arg_name}={kv[0].strip()}"
        else:
            arg_name = arg
            api_arg_str = f"{arg_name}={arg_name}"

        if arg_name == "output_mode":
            api_args.append("output_mode=self._OUTPUT_MODE")
            continue

        api_args.append(api_arg_str)

        arg_str = arg_name
        if arg_name in args_specs:
            spec = args_specs[arg_name]
            spec_str = ""
            if len(spec) > 0:
                spec_str = f": {spec[0]}"

            if len(spec) > 1 and spec[1] == "(optional)":
                spec_str = f": Optional[{spec[0]}] = None"
            arg_str += spec_str

        if arg_str != arg_name:
            method_kwargs.append(arg_str)
        else:
            method_args.append(arg_str)

    method_args_str = ", ".join(method_args)
    if method_kwargs:
        method_args_str += ", " + ", ".join(method_kwargs)
    api_args_str = ", ".join(api_args)

    return api_args_str, method_args_str

    # method_template = get_jinja_env().get_template(
    #     "splunk_client_class_method.tmpl"
    # )


def extract_methods(
    *, method_template: Template, samples: List[str], ta_api_prefix: str
) -> List[str]:
    # method_specs: Dict[str, Dict[str, List[str]]] = {}
    methods: List[str] = []
    for sample in samples:
        args_specs: Dict[str, List[str]] = {}
        for line in sample.splitlines():
            line = line.strip()
            if not line:
                continue
            res = line.split("=", 1)
            if len(res) < 2:
                continue
            arg_name = res[0].strip()
            if arg_name != "api_response":
                args_specs[arg_name] = parse_arg_descriptor(res[1])
            else:  # if arg_name == "api_response":
                call = res[1].strip()
                if call.startswith("api_instance."):
                    call = call[13:].strip()

                api_name = call[: call.find("(")]

                api_name_no_prefix = remove_prefix(
                    api_name=api_name, ta_api_prefix=ta_api_prefix
                )
                method_name, method_type = make_method_name(
                    # api_name_no_prefix
                    api_name_no_prefix=api_name_no_prefix,
                    args_specs=args_specs,
                )

                fst = call.find("(") + 1
                lst = call.rfind(")")
                call_args = call[fst:lst].split(",")
                api_args_str, method_args_str = parse_args(
                    call_args=call_args, args_specs=args_specs
                )

                method_str = method_template.render(
                    method_type=method_type,
                    method_name=method_name,
                    method_args=method_args_str,
                    api_name=api_name,
                    api_args=api_args_str,
                )
                methods.append(method_str)
                # method_specs[method_str] = args_specs

                # self.args_specs = {}
                break
    # return method_specs
    return methods


def noqaE501check(block: str) -> List[str]:
    content: List[str] = []
    for line in block.split("\n"):
        if len(line) > 79:
            line += "  # noqa: E501"
        content.append(line)
    return content


def write_splunk_client(
    *, splunk_client_py: Path, splunk_client_content: str, methods: List[str]
) -> None:
    content = []
    content += noqaE501check("# fmt: off\n")
    content += noqaE501check(splunk_client_content)
    for method in methods:
        content += noqaE501check(method)
    content += noqaE501check("# fmt: on\n")

    with open(splunk_client_py, "w") as file:
        file.write("\n".join(content))


def write_other_classes(*, unified_tests_root_dir: Path) -> None:
    schema = {
        unified_tests_root_dir.parent: [
            {
                "file": "__init__.py",
                "overwrite": False,
            }
        ],
        unified_tests_root_dir: [
            {
                "file": "__init__.py",
                "template": "unified_tests_init.tmpl",
                "overwrite": False,
            },
            {
                "file": "test_settings.py",
                "template": "test_settings.tmpl",
                "overwrite": False,
            },
            {
                "file": "defaults.py",
                "template": "defaults.tmpl",
                "overwrite": False,
            },
            {
                "file": "README.md",
                "template": "README.tmpl",
                "overwrite": False,
            },
        ],
        unified_tests_root_dir
        / "splunk": [
            {
                "file": "__init__.py",
                "overwrite": False,
            },
            {
                "file": "forges.py",
                "template": "splunk_forges.tmpl",
                "overwrite": False,
            },
            {
                "file": "probes.py",
                "template": "splunk_probes.tmpl",
                "overwrite": False,
            },
        ],
        unified_tests_root_dir
        / "splunk"
        / "client": [
            {
                "file": "__init__.py",
                "template": "splunk_client_init.tmpl",
                "overwrite": False,
            },
            {
                "file": "configuration.py",
                "template": "splunk_configuration_class.tmpl",
                "overwrite": False,
            },
            {
                "file": "client.py",
                "template": "splunk_client_class.tmpl",
                "overwrite": False,
            },
        ],
        unified_tests_root_dir
        / "vendor": [
            {
                "file": "__init__.py",
                "overwrite": False,
            },
            {
                "file": "forges.py",
                "template": "vendor_forges.tmpl",
                "overwrite": False,
            },
            {
                "file": "probes.py",
                "template": "vendor_probes.tmpl",
                "overwrite": False,
            },
        ],
        unified_tests_root_dir
        / "vendor"
        / "client": [
            {
                "file": "__init__.py",
                "template": "vendor_client_init.tmpl",
                "overwrite": False,
            },
            {
                "file": "client.py",
                "template": "vendor_client_class.tmpl",
                "overwrite": False,
            },
            {
                "file": "configuration.py",
                "template": "vendor_configuration_class.tmpl",
                "overwrite": False,
            },
        ],
    }

    for folder_path, folder_files in schema.items():
        folder_path.mkdir(parents=True, exist_ok=True)
        for file_info in folder_files:
            file_name = str(file_info["file"])
            file_path = folder_path / file_name
            if (
                not file_path.exists()
                or file_path.exists()
                and file_info.get("overwrite", False)
            ):
                template = str(file_info.get("template", ""))
                if template:
                    template = get_jinja_env().get_template(template).render()
                with file_path.open(mode="w") as file:
                    file.write(template)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations
import json
import logging
from dataclasses import dataclass

from splunklib.client import Endpoint
from splunklib.binding import ResponseReader

from splunk_add_on_ucc_modinput_test.common.splunk_service_pool import (
    SplunkServicePool,
)

logger = logging.getLogger("ucc-modinput-test")


@dataclass
class SplunkInstanceKVStoreAPI:
    """
    This class provides an interface to interact with the Splunk KV store.
    """

    splunk: SplunkServicePool
    collection_name: str
    record_id: str
    app_name: str
    app_user: str = "nobody"

    def _create_record_path(self) -> str:
        return (
            f"/servicesNS/{self.app_user}/{self.app_name}/storage/"
            f"collections/data/{self.collection_name}/{self.record_id}"
        )

    def get_record_from_collection(self) -> dict[str, str]:
        try:
            record_path = self._create_record_path()
            kvstore_endpoint = Endpoint(self.splunk, record_path)
            response = kvstore_endpoint.get()
            response_body = ResponseReader(response["body"]).read()
            return json.loads(response_body.decode("utf-8"))
        except json.JSONDecodeError as e:
            logger.error(
                f"Failed to decode JSON response for KV store record"
                f" {self.record_id} in collection {self.collection_name}: {e}"
            )
            raise
        except Exception as e:
            logger.error(
                f"Failed to get KV store record {self.record_id} from "
                f"collection {self.collection_name}: {e}"
            )
            raise



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from pathlib import Path
import shutil
from typing import Optional
from importlib_resources import files
from splunk_add_on_ucc_modinput_test import resources
from splunk_add_on_ucc_modinput_test.common import bootstrap
import subprocess
from importlib_metadata import version, PackageNotFoundError

SWAGGER_CODEGEN_CLI_VERSION = "3.0.68"


def get_version() -> str:
    try:
        return version("splunk_add_on_ucc_modinput_test")
    except PackageNotFoundError:
        return "unknown"


def initialize(
    *,
    modinput: Path,
) -> Path:
    src = files(resources).joinpath("ucc_modinput_functional")
    for root, _, fls in os.walk(src):
        dest_path = Path(modinput) / Path(root).relative_to(src)
        if not dest_path.exists():
            dest_path.mkdir(parents=True)
        for file in fls:
            shutil.copy2(Path(root) / file, dest_path)

    init_in_tests = modinput.parent / "__init__.py"
    if not init_in_tests.exists():
        init_in_tests.touch()
    return modinput


def generate_swagger_client(
    *,
    openapi: Path,
    tmp: Path,
    swagger_client: Path,
    platform: Optional[str],
) -> Path:
    RESTAPI_CLIENT = "restapi_client"
    GENERATOR = "generator"
    restapi_client_path = tmp / RESTAPI_CLIENT
    generator_path = tmp / GENERATOR
    restapi_client_path.mkdir()
    generator_path.mkdir()
    shutil.copy(str(openapi), str(tmp))
    shutil.copy(
        str(
            files(resources).joinpath(
                "swagger-codegen-generators/src/main/resources/handlebars/python/api_client.mustache"  # noqa: E501
            )
        ),
        str(generator_path),
    )
    shutil.copy(
        str(
            files(resources).joinpath(
                "swagger-codegen-generators/src/main/resources/handlebars/python/rest.mustache"  # noqa: E501
            )
        ),
        str(generator_path),
    )
    docker_run_command = [
        "docker",
        "run",
        "--rm",
        "-v",
        f"{str(tmp.resolve())}:/local",
    ]
    if platform:
        docker_run_command.extend(["--platform", platform])
    docker_run_command.extend(
        [
            f"swaggerapi/swagger-codegen-cli-v3:{SWAGGER_CODEGEN_CLI_VERSION}",
            "generate",
            "-i",
            f"/local/{openapi.name}",
            "-l",
            "python",
            "-o",
            f"/local/{RESTAPI_CLIENT}",
            "-t",
            f"/local/{GENERATOR}/",
        ]
    )

    subprocess.run(
        docker_run_command,
        check=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    shutil.copytree(
        str(restapi_client_path / "swagger_client"),
        str(swagger_client),
    )
    shutil.copy(str(restapi_client_path / "README.md"), str(swagger_client))

    return swagger_client


def generate_splunk_client(
    *,
    swagger_client_readme_md: Path,
    splunk_client_py: Path,
    rest_root: str,
) -> Path:
    """
    Generate the Splunk client from the swagger client README.md
    :param swagger_client_readme_md: Path to the swagger client README.md
    :param splunk_client_py: Path to client.py
    :param rest_root: The REST root
    :return: Path to the generated Splunk client
    """
    jinja_env = bootstrap.get_jinja_env()
    samples = bootstrap.load_readme_examples(
        swagger_client_readme_md=swagger_client_readme_md
    )
    methods = bootstrap.extract_methods(
        method_template=jinja_env.get_template(
            "managed_splunk_client_class_method.tmpl"
        ),
        samples=samples,
        ta_api_prefix=rest_root,
    )
    bootstrap.write_splunk_client(
        splunk_client_py=splunk_client_py,
        splunk_client_content=jinja_env.get_template(
            "managed_splunk_client_class_header.tmpl"
        ).render(),
        methods=methods,
    )
    return splunk_client_py



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
from typing import Any, Dict, List, Tuple
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.entities.test import (
    FrameworkTest,
)
from splunk_add_on_ucc_modinput_test.functional.manager import (
    dependency_manager,
)
from pytest import Item, Session
from _pytest.mark.structures import Mark

from splunk_add_on_ucc_modinput_test.typing import ExecutableKeyType


def _extract_parametrized_data(pyfuncitem: Item) -> Tuple[str, Any]:
    callspec_params = {}
    if hasattr(pyfuncitem, "callspec"):
        callspec_params = pyfuncitem.callspec.params
    return pyfuncitem.keywords.node.name, callspec_params


def _map_forged_tests_to_pytest_items(
    items: List[Item],
) -> Dict[ExecutableKeyType, Item]:
    forged_tests = {}
    for item in items:
        test = dependency_manager.tests.lookup_by_function(item._obj)
        if test:
            forged_tests[test.key] = item
    return forged_tests


def _collect_parametrized_tests(
    items: List[Item],
) -> Dict[ExecutableKeyType, List[Tuple[str, Any]]]:
    parametrized_tests: Dict[ExecutableKeyType, List[Tuple[str, Any]]] = {}
    for item in items:
        parametrized_markers = [
            marker
            for marker in item.own_markers
            if marker.name == "parametrize"
        ]
        if parametrized_markers:
            test = dependency_manager.tests.lookup_by_function(item._obj)
            if test:
                logger.debug(
                    f"parametrized_item: {item},\n\tvars: {vars(item)},\n\tnode: {item.keywords.node},\n\tnode vars: {vars(item.keywords.node)}\n\tvars(test): {vars(test)}"
                )
                if test.key not in parametrized_tests:
                    parametrized_tests[test.key] = []
                parametrized_tests[test.key].append(
                    _extract_parametrized_data(item)
                )

    logger.debug(f"Collected parametrized_tests: {len(parametrized_tests)}")
    for test_key, param_tests in parametrized_tests.items():
        logger.debug(f"\ttest: {test_key}")
        for test_name, test_kwargs in param_tests:
            logger.debug(
                f"\t\ttest name: {test_name}, test kwargs: {test_kwargs}"
            )
    return parametrized_tests


def _collect_skipped_tests(
    items: List[Item],
) -> List[Tuple[FrameworkTest, List[Mark]]]:
    skip_tests = []
    for item in items:
        skipped_markers = [
            marker for marker in item.own_markers if marker.name == "skip"
        ]
        if skipped_markers:
            pytest_funcname, _ = _extract_parametrized_data(item)
            test = dependency_manager.find_test(item._obj, pytest_funcname)
            logger.debug(f"_collect_skipped_tests: {item} ==>> {test}")
            if test:
                skip_tests.append((test, skipped_markers))
    return skip_tests


def _adjust_test_order(items: List[Item]) -> List[Item]:
    tests = []
    logger.debug("Initial test order:")
    for item in items:
        pytest_funcname, _ = _extract_parametrized_data(item)
        test = dependency_manager.find_test(item._obj, pytest_funcname)
        if test:
            logger.debug(f"Item: {item} -> {test.key}")
            ip_tasks, bs_tasks = dependency_manager.tasks.get_tasks_by_type(
                test.key
            )
            tests.append((item, (int(len(ip_tasks) > 0), len(bs_tasks))))
        else:
            tests.append((item, (-1, 0)))

    sorted_items = sorted(tests, key=lambda v: v[1])
    return [item for item, _ in sorted_items]


def _debug_log_test_order(items: List[Item]) -> None:
    logger.debug("Adjusted test order:")
    for item in items:
        pytest_funcname, _ = _extract_parametrized_data(item)
        test = dependency_manager.find_test(item._obj, pytest_funcname)
        logger.debug(f"{type(item)}, {item} => {vars(item)}")
        if test:
            logger.debug(f"{item} -> {test.key}: {test.forges}")
        else:
            logger.debug(f"NOT FOUND: {item}, {pytest_funcname} ")


def _log_test_order(items: List[Item]) -> None:
    order = "\nTest execution order:\n"
    for index, item in enumerate(items):
        pytest_funcname, _ = _extract_parametrized_data(item)
        test = dependency_manager.find_test(item._obj, pytest_funcname)
        if test:
            order += f"{index}. {'::'.join(test.key)}\n"
            test_tasks = dependency_manager.tasks.get_bootstrap_tasks(test.key)
            for level, tasks in enumerate(test_tasks):
                order += f"\tLevel {level}\n"
                if tasks is not None:
                    for task in tasks:
                        order += f"\t\t{task.forge_full_path}\n"
        else:
            logger.debug(f"NOT FOUND: {item}, {pytest_funcname} ")

    logger.info(order)


def _check_session_terminal_output(session: Session) -> None:
    # Access the terminal reporter to check for collection errors
    terminal_reporter = session.config.pluginmanager.get_plugin(
        "terminalreporter"
    )
    if terminal_reporter and "error" in terminal_reporter.stats:
        errors = terminal_reporter.stats["error"]
        if errors:
            logger.error("Errors occurred during test collection:")
            for error in errors:
                logger.error(error.longrepr)
            pytest.exit(
                "Errors occurred during test collection. Exiting pytest.",
                returncode=1,
            )



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

# mypy: disable-error-code="attr-defined,arg-type"

import time
import pytest
from splunklib.client import Service
from splunklib.client import Job
from splunklib.client import Index
import splunklib.results as results
from splunk_add_on_ucc_modinput_test.common import utils
from .splunk_service_pool import SplunkServicePool
import json
from urllib import request, error
import ssl
import certifi
import logging
import re

logger = logging.getLogger("ucc-modinput-test")

MODINPUT_TEST_SPLUNK_DEDICATED_INDEX = "MODINPUT_TEST_SPLUNK_DEDICATED_INDEX"


class Configuration:
    @staticmethod
    def _get_index_from_classic_instance(
        index_name: str,
        client_service: SplunkServicePool,
        acs_stack: str,
        acs_server: str,
        splunk_token: str,
    ) -> Index:
        url = f"{acs_server}/{acs_stack}/adminconfig/v2/indexes/{index_name}"

        headers = {
            "Authorization": "Bearer " + splunk_token,
            "Content-Type": "application/json",
        }
        req = request.Request(url, headers=headers, method="GET")
        context = ssl.create_default_context(cafile=certifi.where())
        try:
            with request.urlopen(req, context=context) as response:
                if response.status == 200:
                    new_host_start = client_service._host.find(acs_stack)
                    host = client_service._host[new_host_start:]
                    service = SplunkServicePool(
                        host=host,
                        port=client_service._port,
                        username=client_service._username,
                        password=client_service._password,
                    )
                    return Index(
                        service, f"/services/data/indexes/{index_name}"
                    )
        except error.HTTPError as e:
            if e.code == 404:
                idx_not_created_msg = (
                    f"Index {index_name} was not found on stack {acs_stack} "
                    f"controlled by {acs_server}."
                )
            else:
                idx_not_created_msg = (
                    f"Failed to retrieve index {index_name} from stack "
                    f"{acs_stack} controlled by {acs_server}. "
                    f"HTTP error: {e.code}"
                )
            utils.logger.critical(idx_not_created_msg)
            return None

    @staticmethod
    def _get_index(
        index_name: str,
        client_service: SplunkServicePool,
        is_cloud: bool = False,
        acs_stack: str | None = None,
        acs_server: str | None = None,
        splunk_token: str | None = None,
    ) -> Index | None:
        if any(
            i.name == index_name
            for i in client_service.indexes.iter(datatype="all")
        ):
            return client_service.indexes[index_name]
        if is_cloud and not client_service._host.startswith(acs_stack):
            return Configuration._get_index_from_classic_instance(
                index_name,
                client_service,
                acs_stack,
                acs_server,
                splunk_token,
            )
        return None

    @staticmethod
    def _validate_index_name(index_name: str) -> None:
        """
        Validate the index name according to Splunk's naming conventions.
        """
        if not index_name:
            reason = "Index name must not be empty"
            logger.error(reason)
            raise ValueError(reason)
        if not re.fullmatch(r"[a-z0-9_-]+", index_name):
            reason = (
                "Index name must consist of only numbers, "
                "lowercase letters, underscores, and hyphens."
            )
            logger.error(reason)
            raise ValueError(reason)
        if index_name.startswith(("_", "-")):
            reason = "Index name cannot begin with an underscore or hyphen."
            logger.error(reason)
            raise ValueError(reason)

    @staticmethod
    def _cloud_create_index(
        index_name: str,
        *,
        datatype: str,
        acs_stack: str,
        acs_server: str,
        splunk_token: str,
    ) -> None:
        Configuration._validate_index_name(index_name)
        url = f"{acs_server}/{acs_stack}/adminconfig/v2/indexes"
        data = json.dumps(
            {
                "datatype": datatype,
                "maxDataSizeMB": 0,
                "name": index_name,
                "searchableDays": 365,
                "splunkArchivalRetentionDays": 366,
                "totalEventCount": "0",
                "totalRawSizeMB": "0",
            }
        ).encode("utf-8")
        headers = {
            "Authorization": "Bearer " + splunk_token,
            "Content-Type": "application/json",
        }
        idx_not_created_msg = f"Index {index_name} was not created on stack \
            {acs_stack} controlled by {acs_server}"

        req = request.Request(url, data=data, headers=headers, method="POST")

        context = ssl.create_default_context(cafile=certifi.where())

        retries_http_errors = 6
        for attempt_http in range(retries_http_errors):
            try:
                with request.urlopen(req, context=context) as response:
                    if response.status == 202:
                        retries = 25
                        backoff_factor = 1
                        for attempt in range(retries):
                            try:
                                get_req = request.Request(
                                    f"{url}/{index_name}",
                                    headers=headers,
                                    method="GET",
                                )
                                with request.urlopen(
                                    get_req, context=context
                                ) as get_response:
                                    if get_response.status == 200:
                                        return
                            except error.HTTPError as e:
                                if e.code == 404:
                                    time.sleep(backoff_factor * (2**attempt))
                                else:
                                    raise
                        idx_not_created_msg += (
                            " or creation time exceeded timeout"
                        )
                        break
            except error.HTTPError as e:
                # Retry logic for specific HTTP error codes:
                # 424: Failed Dependency - indicates a temporary issue with a required resource. # noqa: E501
                # 503: Service Unavailable - suggests the server is temporarily overloaded or down. # noqa: E501
                if e.code in (424, 503):
                    if attempt_http < retries_http_errors:
                        logger.info(
                            f"HTTP Response status {e.code}, retrying to "
                            f"create index {index_name}... "
                            f"{attempt_http + 1}/{retries_http_errors}"
                        )
                        time.sleep(2**attempt_http)
                        continue
                else:
                    # In case of HTTP errors other than 424 and 503
                    raise

            except Exception as e:
                idx_not_created_msg += f"\nException raised:\n{e}"
                break

        logger.critical(idx_not_created_msg)
        pytest.exit(idx_not_created_msg)

    @staticmethod
    def _enterprise_create_index(
        index_name: str, datatype: str, client_service: Service
    ) -> Index:
        idx_not_created_msg = f"Index {index_name} was not created on \
            instance {client_service.host}"
        try:
            new_index = client_service.indexes.create(
                index_name, datatype=datatype
            )
        except Exception as e:
            reason = f"{idx_not_created_msg}\nException raised:\n{e}"
            logger.critical(reason)
            pytest.exit(reason)
        if new_index:
            return new_index
        else:
            logger.critical(idx_not_created_msg)
            pytest.exit(idx_not_created_msg)

    @staticmethod
    def _create_index(
        index_name: str,
        client_service: SplunkServicePool,
        *,
        datatype: str = "event",
        is_cloud: bool = False,
        acs_stack: str | None = None,
        acs_server: str | None = None,
        splunk_token: str | None = None,
    ) -> Index:
        if Configuration._get_index(
            index_name,
            client_service,
            is_cloud,
            acs_stack,
            acs_server,
            splunk_token,
        ):
            reason = f"Index {index_name} already exists"
            logger.critical(reason)
            pytest.exit(reason)
        if is_cloud:
            Configuration._cloud_create_index(
                index_name,
                datatype=datatype,
                acs_stack=acs_stack,
                acs_server=acs_server,
                splunk_token=splunk_token,
            )
            created_index = Configuration._get_index(
                index_name,
                client_service,
                is_cloud,
                acs_stack,
                acs_server,
                splunk_token,
            )
        else:
            created_index = Configuration._enterprise_create_index(
                index_name,
                datatype,
                client_service,
            )
        logger.debug(f"Index {index_name} has just been created")
        return created_index

    __instances: dict[tuple[str, str, str], Configuration] = {}

    @classmethod
    def collect_host(cls) -> str | None:
        return utils.get_from_environment_variable("MODINPUT_TEST_SPLUNK_HOST")

    @classmethod
    def collect_port(cls) -> str | None:
        return utils.get_from_environment_variable("MODINPUT_TEST_SPLUNK_PORT")

    @classmethod
    def collect_username(cls) -> str | None:
        return utils.get_from_environment_variable(
            "MODINPUT_TEST_SPLUNK_USERNAME"
        )

    @classmethod
    def collect_password(cls) -> str | None:
        return utils.get_from_environment_variable(
            "MODINPUT_TEST_SPLUNK_PASSWORD_BASE64",
            string_function=utils.Base64.decode,
        )

    @classmethod
    def collect_splunk_dedicated_index(cls) -> str | None:
        return utils.get_from_environment_variable(
            MODINPUT_TEST_SPLUNK_DEDICATED_INDEX,
            is_optional=True,
        )

    @classmethod
    def collect_splunk_token(cls, is_optional: bool) -> str | None:
        return utils.get_from_environment_variable(
            "MODINPUT_TEST_SPLUNK_TOKEN_BASE64",
            string_function=utils.Base64.decode,
            is_optional=is_optional,
        )

    @classmethod
    def collect_acs_server(cls, is_optional: bool) -> str | None:
        return utils.get_from_environment_variable(
            "MODINPUT_TEST_ACS_SERVER",
            is_optional=is_optional,
        )

    @classmethod
    def collect_acs_stack(cls, is_optional: bool) -> str | None:
        return utils.get_from_environment_variable(
            "MODINPUT_TEST_ACS_STACK",
            is_optional=is_optional,
        )

    def __new__(cls, *args, **kwargs):  # type: ignore
        host = cls.collect_host()
        port = cls.collect_port()
        username = cls.collect_username()

        if host is None or port is None or username is None:
            raise ValueError("Host, port, and username must not be None")

        connection_key = (host, port, username)
        if connection_key in cls.__instances:
            return cls.__instances[connection_key]

        instance = object.__new__(cls)
        instance._host = host
        instance._port = port
        instance._username = username
        instance._password = cls.collect_password()

        dedicated_index_name = cls.collect_splunk_dedicated_index()

        instance._is_cloud = (
            "splunkcloud.com" in instance._host.lower()  # type: ignore
        )
        create_index_in_cloud = instance._is_cloud and not dedicated_index_name
        instance._token = cls.collect_splunk_token(
            is_optional=not create_index_in_cloud
        )
        instance._acs_server = cls.collect_acs_server(
            is_optional=not create_index_in_cloud
        )
        instance._acs_stack = cls.collect_acs_stack(
            is_optional=not create_index_in_cloud
        )
        instance._service = SplunkServicePool(
            host=instance._host,
            port=instance._port,
            username=instance._username,
            password=instance._password,
        )

        if dedicated_index_name:
            instance._dedicated_index = cls._get_index(
                dedicated_index_name, instance._service
            )
            if not instance._dedicated_index:
                reason = f"Environment variable \
                    {MODINPUT_TEST_SPLUNK_DEDICATED_INDEX} set to \
                        {dedicated_index_name}, but Splunk instance \
                            {instance._host} does not \
                                contain such index. Remove the variable \
                                    or create the index."
                logger.critical(reason)
                pytest.exit(reason)
            logger.debug(
                f"Existing index {dedicated_index_name} will be used for \
                    test in splunk {instance._host}"
            )
        else:
            instance._dedicated_index = instance._create_index(
                f"idx_{utils.Common().sufix}",
                instance._service,
                is_cloud=instance._is_cloud,
                acs_stack=instance._acs_stack,
                acs_server=instance._acs_server,
                splunk_token=instance._token,
            )

        logger.info(
            f"Splunk - host:port and user set to \
                {instance._host}:\
                    {instance._port}, \
                        {instance._username}"
        )

        if instance._dedicated_index is not None:
            logger.info(
                f"Splunk - index \
                    {instance._dedicated_index.name} will be \
                        used for the test run"
            )
        cls.__instances[connection_key] = instance
        return instance

    def __init__(self) -> None:
        pass

    @property
    def host(self) -> str:
        return self._host

    @property
    def is_cloud(self) -> bool:
        return self._is_cloud

    @property
    def port(self) -> int:
        return self._port

    @property
    def username(self) -> str:
        return self._username

    @property
    def password(self) -> str:
        return self._password

    @property
    def token(self) -> str:
        return self._token

    @property
    def acs_server(self) -> str:
        return (
            self._acs_server
            if self._acs_server or not self.is_cloud
            else "https://admin.splunk.com"
        )

    @property
    def acs_stack(self) -> str:
        return self._acs_stack

    @property
    def service(self) -> Service:
        return self._service

    @property
    def dedicated_index(self) -> Index:
        return self._dedicated_index


class SearchState:
    def __init__(self, job: Job):
        self._is_done = job["isDone"] == "1"
        self._done_progress = float(job["doneProgress"]) * 100
        self._scan_count = int(job["scanCount"])
        self._event_count = int(job["eventCount"])
        self._result_count = int(job["resultCount"])
        self._results = (
            [
                result
                for result in results.JSONResultsReader(
                    job.results(output_mode="json")
                )
            ]
            if self._is_done
            else None
        )

    @property
    def result_count(self) -> int:
        return self._result_count

    @property
    def results(self) -> list[object] | None:
        return self._results


def search(*, service: Service, searchquery: str) -> SearchState:
    search_state = None
    kwargs_normalsearch = {"exec_mode": "normal"}
    job = service.jobs.create(searchquery, **kwargs_normalsearch)
    while True:
        while not job.is_ready():
            pass
        if job["isDone"] == "1":
            search_state = SearchState(job)
            break
        time.sleep(1)
    job.cancel()
    return search_state



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# mypy: disable-error-code="arg-type"

import argparse
import sys
import tempfile
from pathlib import Path
from datetime import datetime
from typing import Optional, Sequence
from splunk_add_on_ucc_modinput_test import commands, tools
from splunk_add_on_ucc_modinput_test.common import bootstrap, utils
import shutil


def main(argv: Optional[Sequence[str]] = None) -> int:
    class Platform:
        #   https://docs.docker.com/build/building/multi-platform/
        DEFAULT = None
        SUPPORTED = [
            "windows/amd64",
            "linux/amd64",
            "linux/arm64",
            "linux/arm/v7",
            "linux/arm64/v8",
        ]

        @staticmethod
        def validate(value: Optional[str]) -> Optional[str]:
            if value in {Platform.DEFAULT, *Platform.SUPPORTED}:
                return value
            raise argparse.ArgumentTypeError(
                f"""
                Given platform ({value}) is not supported. Supported platforms are:
                {Platform.SUPPORTED}
                """
            )

    class OpenApiPath:
        DEFAULT = "output/*/appserver/static/openapi.json"

        @staticmethod
        def validate(value: str) -> Path:
            if value != OpenApiPath.DEFAULT:
                p = Path(value)
                if not p.exists():
                    raise argparse.ArgumentTypeError(
                        f"""
                        Given openapi.json path ({value}) does not exist.
                        """
                    )
                return p
            else:
                pl = sorted(Path().glob(OpenApiPath.DEFAULT))
                if len(pl) != 1:
                    raise argparse.ArgumentTypeError(
                        f"""
                        Default path ({OpenApiPath.DEFAULT}) does not work in \
                            this case.
                        It returns {len(pl)} results: {[str(x) for x in pl]}
                        Define path to openapi.json
                        """
                    )
                return pl[0]

    class TmpPath:
        DEFAULT = str(Path(tempfile.gettempdir()) / "modinput")

        @staticmethod
        def validate(value: str) -> Path:
            p = Path(value)
            if p.exists():
                p_resolved = p.resolve()
                target = (
                    p_resolved.parent
                    / "backup"
                    / datetime.now().strftime("%Y%m%d%H%M%S")
                    / p_resolved.name
                )
                target.mkdir(parents=True)
                p.rename(target)
            p.mkdir(parents=True)
            return p

    class ClientCodePath:
        DEFAULT = "."

        @staticmethod
        def validate(value: str) -> Path:
            directory = Path(value)
            if not directory.exists():
                raise argparse.ArgumentTypeError(
                    f"Given directory ({value}) has to exist. Create \
                        {directory.resolve()}"
                )
            return directory

    class ModinputPath:
        DEFAULT = "tests/ucc_modinput_functional"

    class FilePath:
        @staticmethod
        def validate(value: str) -> Path:
            f = Path(value)
            if not f.exists():
                raise argparse.ArgumentTypeError(f"{value} does not exist")
            elif not f.is_file():
                raise argparse.ArgumentTypeError(f"{value} is not a file")
            return f

    argv = argv if argv is not None else sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {commands.get_version()}",
    )
    subparsers = parser.add_subparsers(dest="command")
    gen_parser = subparsers.add_parser(
        "gen", description="Generate python client code from openapi.json"
    )
    init_parser = subparsers.add_parser(
        "init",
        description="Initialize modinput tests. This is one time action.",
    )
    base64encode_parser = subparsers.add_parser(
        "base64encode",
        description="Tool to convert complex string (due to special characters \
            or structure) to base64 string",
    )
    base64decode_parser = subparsers.add_parser(
        "base64decode", description="Tool to decode base64 string"
    )

    _p_args = (
        "-p",
        "--platform",
    )
    _p_kwargs = {
        "type": Platform.validate,
        "help": "--platform flag when running swaggerapi/swagger-codegen-cli-v3 docker image",
        "default": Platform.DEFAULT,
    }
    gen_parser.add_argument(*_p_args, **_p_kwargs)
    init_parser.add_argument(*_p_args, **_p_kwargs)

    _o_args = (
        "-o",
        "--openapi-json",
    )
    _o_kwargs = {
        "type": OpenApiPath.validate,
        "help": "openapi.json path. Client code will be generated from this.",
        "default": OpenApiPath.DEFAULT,
    }
    gen_parser.add_argument(*_o_args, **_o_kwargs)
    init_parser.add_argument(*_o_args, **_o_kwargs)

    _t_args = (
        "-t",
        "--tmp",
    )
    _t_kwargs = {
        "type": TmpPath.validate,
        "help": "Temporary directory, where resources needed for client code \
            creation will be stored",
        "default": TmpPath.DEFAULT,
    }
    gen_parser.add_argument(*_t_args, **_t_kwargs)
    init_parser.add_argument(*_t_args, **_t_kwargs)

    _c_args = (
        "-c",
        "--client-code",
    )
    _c_kwargs = {
        "type": ClientCodePath.validate,
        "help": "Path to client code directory. This is target directory.",
        "default": ClientCodePath.DEFAULT,
    }
    gen_parser.add_argument(*_c_args, **_c_kwargs)
    init_parser.add_argument(*_c_args, **_c_kwargs)

    _m_args = (
        "-m",
        "--modinput",
    )
    _m_kwargs = {
        "help": "Path to functional tests target directory.",
        "default": ModinputPath.DEFAULT,
    }
    gen_parser.add_argument(*_m_args, **_m_kwargs)
    init_parser.add_argument(*_m_args, **_m_kwargs)

    feature_group = gen_parser.add_mutually_exclusive_group()
    feature_group.add_argument(
        "--skip-splunk-client-check",
        action="store_true",
        help="Exisitng splunk client will not be checked aginst consistency with swagger client. WARNING: This may lead to inconsistent state of splunk client.",
    )
    feature_group.add_argument(
        "--force-splunk-client-overwritten",
        action="store_true",
        help="Existing splunk client will be overwritten by new one. WARNING: This may lead to loss of existing splunk client.",
    )

    base64decode_parser.add_argument(
        "-s",
        "--string",
        type=str,
        help="Base64 encoded string.",
        required=True,
    )

    group = base64encode_parser.add_mutually_exclusive_group()
    group.add_argument(
        "-f",
        "--file",
        type=FilePath.validate,
        help="Path to input text file.",
    )
    group.add_argument(
        "-s",
        "--string",
        type=str,
        help="String to be base64 encoded.",
    )

    args = parser.parse_args(argv)

    if args.command == "base64encode":
        print(
            tools.base64encode(
                text_file=args.file,
                string=args.string,
            )
        )
    elif args.command == "base64decode":
        print(tools.base64decode(base64_string=args.string))
    elif args.command in ["gen", "init"]:
        docker_err = tools.is_docker_running()
        if docker_err is not None:
            print(docker_err)
            return 1
        modinput_path = Path(args.modinput)
        if args.command == "init" and modinput_path.exists():
            print(
                f"{modinput_path} already exists. Did you want to run 'gen' command?"
            )
            return 1
        if args.command == "gen" and not modinput_path.exists():
            print(
                f"{modinput_path} does not exist. Run 'init' command first or use --modinput/-m with appropriate location."
            )
            return 1
        swagger_client = args.client_code / "swagger_client"
        if swagger_client.exists():
            shutil.rmtree(swagger_client)
        swagger_client = commands.generate_swagger_client(
            openapi=args.openapi_json,
            tmp=args.tmp,
            swagger_client=swagger_client,
            platform=args.platform,
        )

        if args.command == "init":
            bootstrap.write_other_classes(unified_tests_root_dir=modinput_path)
            splunk_client = commands.generate_splunk_client(
                swagger_client_readme_md=swagger_client / "README.md",
                splunk_client_py=bootstrap.get_splunk_client_path(
                    modinput_path
                ),
                rest_root=tools.get_rest_root(openapi=args.openapi_json),
            )
        elif args.command == "gen":
            current_splunk_client = bootstrap.get_splunk_client_path(
                modinput_path
            )
            splunk_client = commands.generate_splunk_client(
                swagger_client_readme_md=swagger_client / "README.md",
                splunk_client_py=args.tmp / bootstrap.CLIENT_FILE_NAME,
                rest_root=tools.get_rest_root(openapi=args.openapi_json),
            )
            if args.skip_splunk_client_check:
                pass
            elif args.force_splunk_client_overwritten:
                if (
                    current_splunk_client.parent.exists()
                    and current_splunk_client.exists()
                ):
                    backup_path = (
                        current_splunk_client.parent
                        / f"{current_splunk_client.name}_{datetime.now().strftime('%Y%m%d%H%M%S')}_backup"
                    )
                    shutil.move(current_splunk_client, backup_path)
                elif not current_splunk_client.parent.exists():
                    current_splunk_client.parent.mkdir()
                shutil.copy(splunk_client, current_splunk_client)
            elif not current_splunk_client.exists():
                print(
                    f"Current splunk client ({current_splunk_client}) does not exist. Use --force-splunk-client-overwritten to create it."
                )
                return 1
            elif utils.md5(file_path=splunk_client) != utils.md5(
                file_path=bootstrap.get_splunk_client_path(modinput_path)
            ):
                print(
                    f"Existing splunk client ({current_splunk_client}) is outdated. Use --force-splunk-client-overwritten to overwrite it or --skip-splunk-client-check if you want to post."
                )
                return 1
            else:
                """
                Existing splunk client is up to date.
                """
                pass
    else:
        parser.print_help()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Any, Callable, Dict, Generator, Tuple, Union, Optional

ProbeGenType = Generator[float, None, Optional[bool]]
ProbeGenFnType = Callable[..., ProbeGenType]
ProbeRegularFnType = Callable[..., Any]
ProbeFnType = Union[ProbeGenFnType, ProbeRegularFnType]

ForgeGenType = Generator[Any, None, None]
ForgeGenFnType = Callable[..., ForgeGenType]
ForgeRegularFnType = Callable[..., Any]
ForgeFnType = Union[ForgeGenFnType, ForgeRegularFnType]

TestFnType = Callable[..., Any]

ArtifactsType = Dict[str, Any]
ExecutableKeyType = Tuple[str, ...]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import Enum


class ForgeProbe(Enum):
    DEFAULT_INTERVAL = 5
    MAX_INTERVAL = 60
    MIN_INTERVAL = 1
    DEFAILT_WAIT_TIME = 300
    MIN_WAIT_TIME = 60
    MAX_WAIT_TIME = 600


class TasksWait(Enum):
    DEFAULT_BOOTSTRAP_TIMEOUT = 1800
    MIN_BOOTSTRAP_TIMEOUT = 300
    MAX_BOOTSTRAP_TIMEOUT = 3600
    DEFAULT_ATTACHED_TIMEOUT = 600
    MIN_ATTACHED_TIMEOUT = 60
    MAX_ATTACHED_TIMEOUT = 1200
    DEFAULT_CHECK_FREQUENCY = 5
    MIN_CHECK_FREQUENCY = 1
    MAX_CHECK_FREQUENCY = 30


class BuiltInArg(Enum):
    SPLUNK_CLIENT = "splunk_client"
    VENDOR_CLIENT = "vendor_client"
    SESSION_ID = "session_id"
    TEST_ID = "test_id"


class ForgeScope(Enum):
    FUNCTION = "function"
    MODULE = "module"
    SESSION = "session"


class Executor(Enum):
    DEFAULT_THREAD_NUMBER = 10
    MIN_THREAD_NUMBER = 10
    MAX_THREAD_NUMBER = 20



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from splunk_add_on_ucc_modinput_test.typing import (
    ArtifactsType,
    ExecutableKeyType,
    ForgeFnType,
    ProbeFnType,
    TestFnType,
)
import time

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Tuple,
    Optional,
    Type,
    Union,
    Set,
)
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.exceptions import (
    SplTaFwkDependencyExecutionError,
    SplTaFwkWaitForDependenciesTimeout,
)
from splunk_add_on_ucc_modinput_test.functional.constants import (
    ForgeScope,
    BuiltInArg,
)
from splunk_add_on_ucc_modinput_test.functional.entities import (
    TestCollection,
    ForgeCollection,
    TaskCollection,
    FrameworkTest,
    FrameworkForge,
    FrameworkTask,
    TaskSetListType,
)
from splunk_add_on_ucc_modinput_test.functional.executor import (
    FrmwkParallelExecutor,
    FrmwkSequentialExecutor,
)
from splunk_add_on_ucc_modinput_test.functional.splunk import (
    SplunkClientBase,
    SplunkConfigurationBase,
)
from splunk_add_on_ucc_modinput_test.functional.vendor import (
    VendorClientBase,
    VendorConfigurationBase,
)

from splunk_add_on_ucc_modinput_test.functional.common.pytest_config_adapter import (
    PytestConfigAdapter,
)
from splunk_add_on_ucc_modinput_test.functional.common.identifier_factory import (
    create_identifier,
    IdentifierType,
)


class forge:
    def __init__(
        self,
        forge_fn: ForgeFnType,
        *,
        probe: Optional[ProbeFnType] = None,
        scope: Optional[Union[ForgeScope, str]] = None,
        **kwargs: Any,
    ) -> None:
        self.forge_fn = forge_fn
        self.probe = probe
        self.scope = scope.value if isinstance(scope, ForgeScope) else scope
        self.kwargs: ArtifactsType = kwargs


class forges:
    def __init__(
        self,
        *forge_list: forge,
        scope: Optional[Union[ForgeScope, str]] = None,
    ) -> None:
        self.forge_list = forge_list
        self.scope = scope.value if isinstance(scope, ForgeScope) else scope


class TestDependencyManager(PytestConfigAdapter):
    def __init__(self) -> None:
        super().__init__()
        self.tests = TestCollection()
        self.forges = ForgeCollection()
        self.tasks = TaskCollection()
        self.executor: Optional[
            Union[FrmwkSequentialExecutor, FrmwkParallelExecutor]
        ] = None
        self._vendor_clients = {
            BuiltInArg.VENDOR_CLIENT.value: (
                VendorClientBase,
                VendorConfigurationBase,
            )
        }
        self._splunk_clients = {
            BuiltInArg.SPLUNK_CLIENT.value: (
                SplunkClientBase,
                SplunkConfigurationBase,
            )
        }
        self._pytest_config = None
        self._session_id = self.generate_session_id()
        self._global_builtin_args_pool: Dict[
            ExecutableKeyType, ArtifactsType
        ] = {}

    @staticmethod
    def generate_session_id() -> str:
        return create_identifier(
            id_type=IdentifierType.ALPHA, in_uppercase=True
        )

    def set_vendor_client_class(
        self,
        vendor_configuration_class: Type[
            VendorConfigurationBase
        ] = VendorConfigurationBase,
        vendor_client_class: Type[VendorClientBase] = VendorClientBase,
        vendor_class_argument_name: str = BuiltInArg.VENDOR_CLIENT.value,
    ) -> None:
        logger.debug(
            f"set_vendor_client_class: {vendor_client_class}, {vendor_configuration_class}, {vendor_class_argument_name}"
        )
        assert (
            vendor_class_argument_name.isidentifier()
        ), "Custom Splunk client argument name must comply with Python variable name requirements"
        assert issubclass(vendor_client_class, VendorClientBase)
        assert issubclass(vendor_configuration_class, VendorConfigurationBase)
        self._vendor_clients[vendor_class_argument_name] = (
            vendor_client_class,
            vendor_configuration_class,
        )

    def set_splunk_client_class(
        self,
        splunk_configuration_class: Type[
            SplunkConfigurationBase
        ] = SplunkConfigurationBase,
        splunk_client_class: Type[SplunkClientBase] = SplunkClientBase,
        splunk_class_argument_name: str = BuiltInArg.SPLUNK_CLIENT.value,
    ) -> None:
        logger.debug(
            f"set_splunk_client_class: {splunk_configuration_class}, {splunk_client_class}, {splunk_class_argument_name}"
        )
        assert (
            splunk_class_argument_name.isidentifier()
        ), "Custom Splunk client argument name must comply with Python variable name requirements"
        assert issubclass(splunk_client_class, SplunkClientBase)
        assert issubclass(splunk_configuration_class, SplunkConfigurationBase)
        self._splunk_clients[splunk_class_argument_name] = (
            splunk_client_class,
            splunk_configuration_class,
        )

    def create_global_builtin_args(
        self,
    ) -> Dict[str, Union[str, VendorClientBase, SplunkClientBase]]:
        global_builtin_args: Dict[
            str, Union[str, VendorClientBase, SplunkClientBase]
        ] = {
            BuiltInArg.SESSION_ID.value: self.session_id,
        }

        for v_prop, (v_client, v_config) in self._vendor_clients.items():
            if self._pytest_config is not None:
                v_conf_instance = v_config(self._pytest_config)
                global_builtin_args[v_prop] = v_client(v_conf_instance)
            logger.debug(
                f"create_global_builtin_args, vendor: {v_prop}, v_config={v_conf_instance} config_id={id(v_conf_instance)}, v_client: {global_builtin_args[v_prop]}"
            )

        for s_prop, (s_client, s_config) in self._splunk_clients.items():
            if self._pytest_config is not None:
                conf_instance = s_config(self._pytest_config)
                global_builtin_args[s_prop] = s_client(conf_instance)
            logger.debug(
                f"create_global_builtin_args, splunk: {s_prop}, s_config={conf_instance} config_id={id(conf_instance)}, s_client: {global_builtin_args[s_prop]}"
            )

        return global_builtin_args

    def get_global_builtin_args(
        self, test_key: ExecutableKeyType
    ) -> Dict[str, Any]:
        if test_key not in self._global_builtin_args_pool:
            logger.debug(f"create_global_builtin_args for test {test_key}:")
            self._global_builtin_args_pool[
                test_key
            ] = self.create_global_builtin_args()

        return self._global_builtin_args_pool[test_key]

    @property
    def session_id(self) -> str:
        return self._session_id

    def _interpret_scope(
        self, scope: Optional[str], test: FrameworkTest
    ) -> Optional[str]:
        if scope is None:
            return None

        if scope == ForgeScope.FUNCTION.value:
            scope = test.full_path
        elif scope == ForgeScope.MODULE.value:
            scope = test.source_file

        return scope

    def forge_find_or_make(
        self, forge_fn: Callable[..., Any], scope: str, is_bootstrap: bool
    ) -> FrameworkForge:
        frg = FrameworkForge(forge_fn, scope)
        found = self.forges.get(frg.key)
        if not found:
            self.forges.add(frg)
            logger.debug(
                f"BIND created frg {id(frg)} - {frg} - {frg.key}, is_bootstrap: {is_bootstrap}"
            )
        else:
            frg = found
            logger.debug(
                f"BIND found frg {id(frg)} - {frg} - {frg.key}, is_bootstrap: {is_bootstrap}"
            )
        return frg

    def bind(
        self,
        test_fn: Callable[..., Any],
        scope: Optional[str],
        frg_fns: List[forge],
        is_bootstrap: bool,
    ) -> FrameworkTest:
        logger.debug(f"bind: {test_fn} -> {frg_fns}")

        test = self.tests.lookup_by_function(test_fn)
        if not test:
            test = FrameworkTest(test_fn)
            self.tests.add(test)

        frg_group_scope = self._interpret_scope(scope, test)

        frg_list = []
        for f in frg_fns:
            if frg_group_scope is None and f.scope is None:
                frg_scope = ForgeScope.SESSION.value
            elif f.scope is not None:
                tmp_scope = self._interpret_scope(f.scope, test)
                frg_scope = (
                    tmp_scope
                    if tmp_scope is not None
                    else ForgeScope.SESSION.value
                )
            elif frg_group_scope is not None:
                frg_scope = frg_group_scope
            else:
                frg_scope = ForgeScope.SESSION.value

            frg = self.forge_find_or_make(f.forge_fn, frg_scope, is_bootstrap)

            frg_list.append(
                FrameworkTask(test, frg, is_bootstrap, f.kwargs, f.probe, self)
            )

            frg.link_test(test.key)
            test.link_forge(frg.key)

        self.tasks.add(frg_list)

        return test

    def unregister_test(
        self, test_key: ExecutableKeyType
    ) -> Optional[FrameworkTest]:
        test = self.tests.pop(test_key, None)
        if test:
            for frg_key in test.forges:
                frg = self.forges.get(frg_key)
                if frg:
                    frg.unlink_test(test.key)
            logger.debug(
                f"unregister_test: Test {test_key} has been unregistered"
            )

        return test

    def find_test(
        self, test_fn: TestFnType, parametrized_name: str
    ) -> Optional[FrameworkTest]:
        test_obj = FrameworkTest(test_fn, parametrized_name)
        return self.tests.get(test_obj.key)

    def dump_tests(self) -> None:
        logger.debug(f"DUMP TESTS: {len(self.tests.items())}")
        for key, test in self.tests.items():
            logger.debug(f"test key:{key} value: {test}")

    def copy_task_for_parametrized_test(
        self,
        test: FrameworkTest,
        extra_kwargs: ArtifactsType,
        src_task: FrameworkTask,
        is_function_scope: bool,
    ) -> FrameworkTask:
        if is_function_scope:
            frg = self.forge_find_or_make(
                forge_fn=src_task._forge._function,
                scope=test.full_path,
                is_bootstrap=src_task.is_bootstrap,
            )
        else:
            frg = src_task._forge

        frg.link_test(test.key)
        test.link_forge(frg.key)
        probe = src_task.get_probe_fn()
        is_bootstrap = src_task.is_bootstrap
        kwargs = src_task.get_forge_kwargs_copy()
        kwargs.update(extra_kwargs)
        return FrameworkTask(test, frg, is_bootstrap, kwargs, probe, self)

    def expand_parametrized_tests(
        self,
        parametrized_tests: Dict[ExecutableKeyType, List[Tuple[str, Any]]],
    ) -> None:
        for test_key, param_tests in parametrized_tests.items():
            test = self.unregister_test(test_key)
            if test is None:
                logger.debug(f"TEST NOT FOUND: {test_key}")
                continue

            logger.debug(f"Test found: {test.key}")
            test_tasks = self.tasks.remove_test_tasks(test_key)
            assert test_tasks
            for test_name, parametrized_kwargs in param_tests:
                logger.debug(
                    f"adding test: {test.key} => test_name: {test_name}, kwargs: {parametrized_kwargs}"
                )
                parametrized_test = FrameworkTest(test._function, test_name)
                logger.debug(
                    f"create parametrized test: {parametrized_test.key}, {vars(parametrized_test)}"
                )
                self.tests.add(parametrized_test)

                for parallel_tasks in test_tasks:
                    frg_list = []
                    if parallel_tasks is not None:
                        for src_task in parallel_tasks:
                            is_function_scope = (
                                src_task._forge.scope == test.full_path
                            )
                            parametrized_task = (
                                self.copy_task_for_parametrized_test(
                                    parametrized_test,
                                    parametrized_kwargs,
                                    src_task,
                                    is_function_scope,
                                )
                            )
                            frg_list.append(parametrized_task)

                    self.tasks.add(frg_list)

                    logger.debug(
                        f"parametrized_test.link_forge: {parametrized_test.key}: {frg_list} => {parametrized_test}"
                    )

    def _log_dep_exec_matrix(
        self, tests: List[FrameworkTest], dep_mtx: List[TaskSetListType]
    ) -> None:
        matrix = "\nBootstrap Dependency execution matrix:\n"
        for step_index, group in enumerate(dep_mtx):
            matrix += f"Step {step_index+1}:\n"
            for test_index, test_tasks in enumerate(group):
                matrix += f"\ttest {test_index+1}: {'::'.join(tests[test_index].key)}\n"
                if test_tasks is not None:
                    for task in test_tasks:
                        matrix += f"\t\tDependency {task.forge_full_path}, scope {task.forge_scope}\n"
                else:
                    matrix += "\t\tNo dependencies at this step\n"
        logger.info(matrix)

    def remove_skipped_tests(
        self, skipped_tests_keys: List[Tuple[str, ...]]
    ) -> None:
        for test_key in skipped_tests_keys:
            self.unregister_test(test_key)

    def synch_tests_with_pytest_list(
        self, pytest_test_set_keys: Set[ExecutableKeyType]
    ) -> None:
        tests_to_remove = [
            test_key
            for test_key in self.tests.keys()
            if test_key not in pytest_test_set_keys
        ]
        self.remove_skipped_tests(tests_to_remove)

    def build_bootstrap_matrix(self) -> List[TaskSetListType]:
        tests = list(self.tests.values())

        exec_steps = []
        step_index = 0
        while True:
            step_tasks: List[Optional[List[FrameworkTask]]] = [None] * len(
                tests
            )
            for pos, test in enumerate(tests):
                tasks = self.tasks.get_bootstrap_tasks(test.key)
                if step_index < len(tasks):
                    step_tasks[pos] = tasks[step_index]
            if not any(step_tasks):
                break
            exec_steps.append(step_tasks)
            step_index += 1

        self._log_dep_exec_matrix(tests, exec_steps)
        return exec_steps

    def start_bootstrap_execution(self) -> None:
        if self.tests.is_empty or self.tasks.is_empty:
            return

        logger.info("Starting bootstrap forges execution.")
        if self.executor is None:
            if self.sequential_execution:
                self.executor = FrmwkSequentialExecutor(self)
            else:
                self.executor = FrmwkParallelExecutor(
                    self, self.number_of_threads
                )

        deps_exec_mtx = dependency_manager.build_bootstrap_matrix()
        if deps_exec_mtx:
            self._execution_timeout = time.time() + self.bootstrap_wait_timeout
            self.executor.start(deps_exec_mtx)

    def inplace_tasks_execution(
        self, deps_exec_mtx: List[TaskSetListType]
    ) -> None:
        logger.debug(f"start inplace_tasks_execution:{deps_exec_mtx}")
        if not deps_exec_mtx:
            return
        assert self.executor is not None
        logger.debug("inplace_tasks_execution is about to start")
        self._execution_timeout = (
            time.time() + self.attached_tasks_wait_timeout
        )
        self.executor.start(deps_exec_mtx)
        self.executor.wait(is_bootstrap=False)

    def shutdown(self) -> None:
        logger.info("Shutting down dependency manager...")

        if self.executor is not None:
            self.executor.shutdown()
            self.executor = None

    def check_all_tests_executed(self) -> bool:
        executed = [test.is_executed for test in self.tests.values()]
        return all(executed)

    def check_tests_executed(
        self, tests_keys: List[ExecutableKeyType]
    ) -> bool:
        executed = []
        for test_key in tests_keys:
            test = self.tests.get(test_key)
            if test is not None:
                executed.append(test.is_executed)
        return all(executed)

    def try_to_unblock_inplace_teardowns(self, test: FrameworkTest) -> None:
        inplace_tasks = self.tasks.enumerate_inplace_tasks(test.key)
        for _, _, task in inplace_tasks:
            if self.check_tests_executed(task.forge_test_keys):
                task.unblock_forge_teardown()

    def teardown_test_dependencies(self, test: FrameworkTest) -> None:
        self.try_to_unblock_inplace_teardowns(test)
        tasks = list(self.tasks.enumerate_tasks(test.key))
        for _, _, task in reversed(tasks):
            task.teardown()

    def teardown_test(self, test: FrameworkTest) -> None:
        logger.debug(f"teardown test:{test}")
        test.mark_executed()
        self.teardown_test_dependencies(test)

    def _check_failed_tasks(
        self, test: FrameworkTest, done_tasks: List[FrameworkTask]
    ) -> None:
        failed_tasks = [
            task.forge_key for task in done_tasks if task._setup_errors
        ]
        logger.debug(
            f"DONE TASKS for {test.key}: {[task.forge_key for task in done_tasks]}"
        )
        logger.debug(f"FAILED TASKS for {test.key}: {failed_tasks}")
        for task in done_tasks:
            if task.setup_failed:
                test_path = test.full_path
                msg = f"Dependency {task.forge_full_path} failed to execute for test {test_path} with error:\n{task.setup_error}"
                logger.error(msg)
                raise SplTaFwkDependencyExecutionError(msg)

    def _report_timeout(
        self, test: FrameworkTest, pending_tasks: List[FrameworkTask]
    ) -> None:
        msg = f"{test} exceeded {self.bootstrap_wait_timeout} seconds timeout while waiting for dependencies:"
        for task in pending_tasks:
            msg += f"\n\t{task.forge_full_path}, self id: {id(task)}, scope: {task.forge_scope}, exec_id: {task._exec_id} is_executed: {task.is_executed}, is_failed: {task.setup_failed}"
        logger.error(msg)
        raise SplTaFwkWaitForDependenciesTimeout(msg)

    def wait_for_test_bootstrap(self, test: FrameworkTest) -> None:
        while True:
            done, pending = self.tasks.bootstrap_tasks_by_state(test.key)
            self._check_failed_tasks(test, done)

            if not pending:
                break
            elif time.time() > self._execution_timeout:
                self._report_timeout(test, pending)

            logger.debug(f"{test} is waiting for bootstrap dependencies")
            time.sleep(self.completion_check_frequency)

        logger.debug(f"{test} bootstrap dependencies are ready")

    def execute_test_inplace_forges(self, test: FrameworkTest) -> None:
        logger.debug(f"Executing attached forges for test {test}")
        exec_steps = []
        inplace_tasks = self.tasks.get_inplace_tasks(test.key)
        dump = (
            f"Execution matrix for attached forges for {test}: {inplace_tasks}"
        )
        for tasks in inplace_tasks:
            dump += f"\t{[tasks]}"
            exec_steps.append([tasks])
        logger.debug(dump)
        self.inplace_tasks_execution(exec_steps)

    def test_setup_error_report(
        self, test: FrameworkTest
    ) -> Generator[Tuple[FrameworkTask, str], None, None]:
        for _, _, task in self.tasks.enumerate_tasks(test.key):
            if task.setup_failed:
                yield task, task.setup_error

    def test_teardown_error_report(
        self, test: FrameworkTest
    ) -> Generator[Tuple[FrameworkTask, str], None, None]:
        for _, _, task in self.tasks.enumerate_tasks(test.key):
            if task.teardown_failed:
                yield task, task.teardown_error

    def test_error_report(
        self, test: FrameworkTest
    ) -> Generator[Tuple[FrameworkTask, str], None, None]:
        for _, _, task in self.tasks.enumerate_tasks(test.key):
            if task.failed:
                yield task, task.error


dependency_manager = TestDependencyManager()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from copy import deepcopy
from typing import Any, Dict, Optional, Set
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.constants import BuiltInArg
from splunk_add_on_ucc_modinput_test.functional.entities.executable import (
    ExecutableBase,
)
from splunk_add_on_ucc_modinput_test.functional.common.identifier_factory import (
    create_identifier,
    IdentifierType,
)
from splunk_add_on_ucc_modinput_test.typing import (
    ExecutableKeyType,
    TestFnType,
)


class FrameworkTest(ExecutableBase):
    def __init__(
        self,
        function: TestFnType,
        altered_name: Optional[str] = None,
    ) -> None:
        super().__init__(function)
        self.forges: Set[ExecutableKeyType] = set()
        self._is_executed: bool = False
        self._artifacts: Dict[str, Any] = {}
        if altered_name:
            self._fn_name = altered_name
        self._test_id = self.generate_test_id()

    @staticmethod
    def generate_test_id() -> str:
        return create_identifier(
            id_type=IdentifierType.ALPHA, in_uppercase=True
        )

    @property
    def test_id(self) -> str:
        return self._test_id

    @property
    def is_executed(self) -> bool:
        return self._is_executed

    @property
    def artifacts(self) -> Dict[str, Any]:
        return self._artifacts

    @property
    def name(self) -> str:
        return self.key[1]

    @property
    def path(self) -> str:
        return self.source_file

    @property
    def full_path(self) -> str:
        return "::".join(self.key)

    @property
    def original_full_path(self) -> str:
        return "::".join(self.original_key)

    def update_artifacts(self, artifacts: Dict[str, Any]) -> None:
        assert isinstance(artifacts, dict)
        self._artifacts.update(artifacts)

    def mark_executed(self) -> None:
        logger.debug(f"TEST: mark_executed {self}")
        self._is_executed = True

    def link_forge(self, forge_key: ExecutableKeyType) -> None:
        assert (
            forge_key not in self.forges
        ), "Attempt to assign the same forge multiply times or duplicated test name"
        self.forges.add(forge_key)

    def __repr__(self) -> str:
        return f"<Test {'::'.join(self.key)}>"

    @property
    def builtin_args(self) -> Dict[str, str]:
        return {BuiltInArg.TEST_ID.value: self.test_id}

    @property
    def artifacts_copy(self) -> Dict[str, Any]:
        try:
            # make a copy of artifacts if possible
            return deepcopy(self._artifacts)
        except TypeError:
            # copy is not possible, get artifacts by reference
            logger.warning(
                f"deepcopy of test artifacts is not possible, returning them by reference.\n\test:{self.key}\n\ttest artifacts: {self._artifacts}"
            )
            return self._artifacts

    def collect_required_kwargs(
        self, global_builtin_args: Dict[str, str]
    ) -> Dict[str, Any]:
        all_args = self.artifacts_copy
        all_args.update(self.builtin_args)
        all_args.update(global_builtin_args)
        required_args = {
            k: v for k, v in all_args.items() if k in self._required_args
        }
        logger.debug(
            f"Finish collect_required_kwargs for test {self.key}: kwargs={required_args}"
        )
        return required_args



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations
from typing import TYPE_CHECKING

from splunk_add_on_ucc_modinput_test.typing import (
    ArtifactsType,
    ForgeGenType,
    ProbeGenType,
)

if TYPE_CHECKING:
    from splunk_add_on_ucc_modinput_test.typing import (
        ProbeFnType,
        ProbeGenFnType,
        ExecutableKeyType,
    )

import inspect
import time
import types
import random
import traceback
from copy import deepcopy
from typing import Any, Generator, Optional, List
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.common.pytest_config_adapter import (
    PytestConfigAdapter,
)
from splunk_add_on_ucc_modinput_test.functional.constants import ForgeProbe

from splunk_add_on_ucc_modinput_test.functional.entities.forge import (
    FrameworkForge,
)
from splunk_add_on_ucc_modinput_test.functional.entities.test import (
    FrameworkTest,
)
from splunk_add_on_ucc_modinput_test.functional.exceptions import (
    SplTaFwkWaitForProbeTimeout,
)
from splunk_add_on_ucc_modinput_test.functional.entities.executable import (
    ExecutableBase,
)
from splunk_add_on_ucc_modinput_test.functional.splunk import SplunkClientBase
from splunk_add_on_ucc_modinput_test.functional.vendor import VendorClientBase


class FrameworkTask:
    def __init__(
        self,
        test: FrameworkTest,
        forge: FrameworkForge,
        is_bootstrap: bool,
        forge_kwargs: dict[str, Any],
        probe_fn: ProbeFnType | None,
        config: PytestConfigAdapter,
    ):
        self._config = config
        self._test = test
        self._forge = forge
        self._is_bootstrap = is_bootstrap
        self._forge_initial_kwargs = forge_kwargs
        self._exec_id: str | None = None
        self._is_executed = False
        self._teardown: Generator[None, None, None] | None = None
        self._setup_errors: list[str] = []
        self._teardown_errors: list[str] = []
        self._result: object | None = None
        self._global_builtin_args: dict[str, Any] = {}
        self._forge_kwargs: dict[str, Any] = {}
        self._probe: ExecutableBase | None = None
        self._probe_fn: ProbeFnType | None = None
        self._probe_gen: ProbeGenFnType | None = None
        self._probe_kwargs: dict[str, Any] = {}
        self.apply_probe(probe_fn)

    def __repr__(self) -> str:
        return f"{id(self)} - {super().__repr__()}, is_executed={self.is_executed}, is_bootstrap={self._is_bootstrap}, dep: {id(self._forge)} - {self._forge} - {self.forge_key}"

    @property
    def is_bootstrap(self) -> bool:
        return self._is_bootstrap

    @property
    def is_executed(self) -> bool:
        return self._is_executed

    @property
    def error(self) -> str:
        errors = self._setup_errors + self._teardown_errors
        return "\n".join(errors)

    @property
    def setup_error(self) -> str:
        return "\n".join(self._setup_errors)

    @property
    def teardown_error(self) -> str:
        return "\n".join(self._teardown_errors)

    @property
    def setup_failed(self) -> bool:
        return bool(self._setup_errors)

    @property
    def teardown_failed(self) -> bool:
        return bool(self._teardown_errors)

    @property
    def failed(self) -> bool:
        return self.setup_failed or self.teardown_failed

    @property
    def result(self) -> object:
        return self._result

    @property
    def forge_key(self) -> tuple[str, ...]:
        return self._forge.key

    @property
    def forge_scope(self) -> str:
        return self._forge.scope

    @property
    def forge_test_keys(self) -> list[ExecutableKeyType]:
        return list(self._forge.tests_keys)

    @property
    def forge_name(self) -> str:
        return self.forge_key[1]

    @property
    def forge_path(self) -> str:
        return self.forge_key[0]

    @property
    def forge_full_path(self) -> str:
        return "::".join(self.forge_key[:2])

    @property
    def test_key(self) -> tuple[str, ...]:
        return self._test.key

    @property
    def test_name(self) -> str:
        return self._test.name

    @property
    def test_path(self) -> str:
        return self._test.path

    @property
    def test_full_path(self) -> str:
        return self._test.full_path

    @property
    def probe_name(self) -> str | None:
        if self._probe:
            return self._probe.key[1]
        return None

    @property
    def probe_path(self) -> str | None:
        if self._probe:
            return self._probe.key[0]
        return None

    @property
    def probe_full_path(self) -> str | None:
        if self._probe:
            return "::".join(self._probe.key)
        return None

    @property
    def summary(self) -> str:
        return (
            f"\ntest: {self.test_name},"
            + f"\n\tlocation: {self.test_path},"
            + f"\n\tforge: {self.forge_name},"
            + f"\n\t\tlocation: {self.forge_path},"
            + f"\n\t\tscope: {self.forge_scope},"
            + f"\n\t\texec id: {self._exec_id},"
            + f"\n\t\tkwargs: {self._forge_kwargs},"
            + f"\n\tprobe: {self.probe_name},"
            + f"\n\t\tlocation: {self.probe_path},"
            + f"\n\t\tkwargs: {self._probe_kwargs},"
        )

    @property
    def default_artifact_name(self) -> str:
        return self._forge.original_name

    def block_forge_teardown(self) -> None:
        logger.debug(f"BLOCK teardown for forge {self._forge.key}")
        self._forge.block_teardown()

    def unblock_forge_teardown(self) -> None:
        logger.debug(f"UNBLOCK teardown for forge {self._forge.key}")
        self._forge.unblock_teardown()

    def make_kwarg(self, test_result: object | None) -> ArtifactsType:
        if test_result is None:
            return {}
        elif not isinstance(test_result, dict):
            return {self.default_artifact_name: test_result}
        else:
            return test_result

    def apply_probe(self, probe_fn: ProbeFnType | None) -> None:
        self._probe_fn = probe_fn
        if callable(probe_fn):
            self._probe = ExecutableBase(probe_fn)

        if inspect.isgeneratorfunction(self._probe_fn):
            self._probe_gen = probe_fn
        elif callable(self._probe_fn):

            def _probe_default_gen(
                **probe_args: Any,
            ) -> Generator[int, None, bool]:
                if probe_fn is not None:
                    while not probe_fn(**probe_args):
                        yield self._config.probe_invoke_interval
                return True

            self._probe_gen = _probe_default_gen
        else:
            self._probe_gen = None

        if self._probe_gen and self._probe_fn is not None:
            sig = inspect.signature(self._probe_fn)
            self._probe_required_args = list(sig.parameters.keys())
        else:
            self._probe_required_args = []

    def collect_available_kwargs(self) -> dict[str, Any]:
        available_kwargs = self._test.artifacts_copy
        available_kwargs.update(self.get_forge_kwargs_copy())
        available_kwargs.update(self._global_builtin_args)
        available_kwargs.update(self._test.builtin_args)
        return available_kwargs

    def prepare_forge_call_args(
        self, global_builtin_args: dict[str, Any]
    ) -> None:
        logger.debug(f"EXECTASK: prepare_forge_call_args {self}")

        self._global_builtin_args = global_builtin_args

        available_kwargs = self.collect_available_kwargs()
        self._forge_kwargs = self._forge.filter_requied_kwargs(
            available_kwargs
        )

        logger.debug(
            f"EXECTASK: prepare_forge_call_args for {self.forge_key}:\n\ttest required args: {self._test.required_args_names}\n\ttest artifacts: {self._test.artifacts}\n\tforge initial kwargs: {self._forge_initial_kwargs}\n\tforge kwargs: {self._forge_kwargs}\n\ttask available kwargs: {available_kwargs}"
        )

    def _get_comparable_args(self) -> dict[str, Any]:
        return {
            k: v
            for k, v in self._forge_kwargs.items()
            if not isinstance(v, (SplunkClientBase, VendorClientBase))
        }

    def get_forge_kwargs_copy(self) -> dict[str, Any]:
        try:
            # make a copy of initial arguments if possible
            return deepcopy(self._forge_initial_kwargs)
        except TypeError:
            # copy is not possible, get initial arguments by reference
            logger.warning(
                f"deepcopy of forge initial arguments is not possible, returning them by reference.{self.summary}"
            )
            return self._forge_initial_kwargs

    def get_probe_fn(self) -> ProbeFnType | None:
        return self._probe_fn

    def invoke_probe(self) -> ProbeGenType:
        if callable(self._probe_gen):
            result = yield from self._probe_gen(**self._probe_kwargs)
            return result
        return None

    def prepare_probe_kwargs(self, extra_args: dict[str, Any] = {}) -> None:
        available_kwargs = self.collect_available_kwargs()
        available_kwargs.update(extra_args)

        self._probe_kwargs = {
            k: v
            for k, v in available_kwargs.items()
            if k in self._probe_required_args
        }

    def wait_for_probe(self, last_result: ArtifactsType) -> bool | None:
        logger.debug(
            f"WAIT FOR PROBE started\n\ttest {self.test_key}\n\tforge {self.forge_key}\n\tprobe {self._probe_fn}"
        )
        if not self._probe_gen:
            return None

        probe_start_time = time.time()
        self.prepare_probe_kwargs(last_result)
        expire_time = time.time() + self._config.probe_wait_timeout
        logger.debug(
            f"WAIT FOR PROBE\n\ttest {self.test_key}\n\tforge {self.forge_key}\n\tprobe {self._probe_fn}\n\tprobe_gen {self._probe_gen}\n\tprobe_args {self._probe_kwargs}"
        )

        result = None
        try:
            it = self.invoke_probe()
            while True:
                interval = next(it)
                if time.time() > expire_time:
                    result = False
                    msg = f"Test {self.test_key}, forge {self.forge_key}: probe {self._probe_fn} exceeded {self._config.probe_wait_timeout} seconds timeout"
                    raise SplTaFwkWaitForProbeTimeout(msg)

                if not isinstance(interval, int):
                    interval = self._config.probe_invoke_interval
                elif interval > ForgeProbe.MAX_INTERVAL.value:
                    interval = ForgeProbe.MAX_INTERVAL.value
                elif interval < ForgeProbe.MIN_INTERVAL.value:
                    interval = ForgeProbe.MIN_INTERVAL.value
                time.sleep(interval)
        except StopIteration as sie:
            result = sie.value

        logger.info(
            f"Forge probe has finished execution, result: {result}, time taken {time.time() - probe_start_time} seconds:{self.summary}"
        )

        return result

    def mark_as_failed(self, error: str | Exception, prefix: str) -> None:
        if isinstance(error, Exception):
            traceback_info = traceback.format_exc()
            report = f"{prefix}: {error}{self.summary}\n{traceback_info}"
        else:
            report = f"{prefix}: {error}{self.summary}"
        logger.error(report)
        self._setup_errors.append(report)
        self._is_executed = True

    def mark_as_executed(self) -> None:
        self._is_executed = True
        logger.debug(
            f"MARK TASK EXECUTED: {self.forge_full_path},\n\tself id: {id(self)},\n\tscope: {self.forge_scope},\n\texec_id: {self._exec_id},\n\ttest: {self.test_key},\n\tis_executed: {self.is_executed},\n\tis_failed: {self.failed},\n\terrors: {self._setup_errors}"
        )

    def _save_generator_teardown(self, gen: ForgeGenType | None) -> None:
        self._teardown = gen

    def _save_class_teardown(self) -> None:
        if not isinstance(self._forge._function, types.FunctionType):
            attr = getattr(self._forge._function, "teardown", None)
            if callable(attr):
                self._teardown = attr

    def update_test_artifacts(self, artifacts: dict[str, Any]) -> None:
        self._test.update_artifacts(artifacts)

    @staticmethod
    def same_args(args1: Any, args2: Any) -> bool:
        if type(args1) != type(args2):
            return False

        if isinstance(args1, (list, tuple)):
            if len(args1) != len(args2):
                return False
            for arg1, arg2 in zip(args1, args2):
                if not FrameworkTask.same_args(arg1, arg2):
                    return False
            return True

        if isinstance(args1, dict):
            if len(args1) != len(args2):
                return False
            if set(args1.keys()).difference(set(args1.keys())):
                return False
            if set(args2.keys()).difference(set(args1.keys())):
                return False
            for k, v in args1.items():
                if not FrameworkTask.same_args(v, args2[k]):
                    return False
            return True

        return args1 == args2

    def same_tasks(self, other_task: FrameworkTask) -> bool:
        if self.forge_key != other_task.forge_key:
            return False

        args1 = self._get_comparable_args()
        args2 = other_task._get_comparable_args()
        return FrameworkTask.same_args(args1, args2)

    def reuse_forge_execution(
        self, exec_id: str, result: Any, errors: list[str]
    ) -> None:
        logger.debug(
            f"reuse execution {exec_id}:\n\tTask: {self.test_key}\n\tDep: {self.forge_key}\n\tresult: {result}\n\terrors: {errors}"
        )
        self._forge.reuse_execution(exec_id)
        self._exec_id = exec_id
        self._result = result
        self._setup_errors = errors

    def use_previous_executions(
        self, args_to_match: ArtifactsType
    ) -> tuple[bool, ArtifactsType]:
        logger.debug(
            f"Dep {self.forge_key}: look for {self._forge_kwargs} in {self._forge.executions}"
        )
        for prev_exec in self._forge.executions:
            logger.debug(
                f"EXECTASK: COMPARE ARGS:\n\tprev exec: {prev_exec.kwargs}\n\tcurrent args {args_to_match}"
            )
            if self.same_args(prev_exec.kwargs, args_to_match):
                self.reuse_forge_execution(
                    prev_exec.id, prev_exec.result, prev_exec.errors
                )
                logger.info(f"Forge execution has been REUSED:{self.summary}")
                return True, prev_exec.result
        return False, {}

    def execute(self) -> None:
        logger.debug(
            f"EXECTASK: execute {self}:\n\texecutions {self._forge.executions}\n\tdep_kwargs: {self._forge_kwargs}\n\tprobe {self._probe_fn}"
        )
        comp_kwargs = self._get_comparable_args()
        reuse, result = self.use_previous_executions(comp_kwargs)
        if not reuse:
            self._exec_id = (
                f"{int(time.time()*1000000)}{random.randint(0, 1000):03}"
            )
            logger.debug(
                f"\nEXECTASK self id: {id(self)}: execute {self} - similar executions not found,\n\tTEST: {self.test_key},\n\tself._required_args: {self._forge._required_args},\n\tself._forge_initial_kwargs: {self._forge_initial_kwargs},\n\tcall_args: {self._forge_kwargs},\n\ttest artifacts: {self._test.artifacts}"
            )
            try:
                forge_start_time = time.time()
                if self._forge._is_generatorfunction:
                    logger.debug(
                        f"EXECTASK: dependency {self._forge} is a generator function"
                    )
                    it = self._forge._function(**self._forge_kwargs)
                    try:
                        result = self.make_kwarg(next(it))
                        self._save_generator_teardown(it)
                    except StopIteration as sie:
                        result = self.make_kwarg(sie.value)
                else:
                    result = self.make_kwarg(
                        self._forge._function(**self._forge_kwargs)
                    )
                    self._save_class_teardown()
                logger.info(
                    f"Forge has been executed successfully, time taken {time.time() - forge_start_time} seconds:{self.summary}"
                )
            except Exception as e:
                traceback_info = traceback.format_exc()
                report = f"Forge has failed to execute: {e}{self.summary}\n{traceback_info}"
                logger.error(report)
                self._setup_errors.append(report)

        try:
            if not self.setup_failed:
                probe_res = self.wait_for_probe(result)
                probe_fn = self.get_probe_fn()
                if probe_res is not None and probe_fn is not None:
                    result[probe_fn.__name__] = probe_res
        except Exception as e:
            traceback_info = traceback.format_exc()
            report = f"Forge probe has failed to execute: {e}{self.summary}\n{traceback_info}"
            logger.error(report)
            self._setup_errors.append(report)

        if not reuse:
            self._result = result
            assert (
                self._exec_id is not None
            ), "_exec_id must be assigned earlier in this method"
            self._forge.register_execution(
                self._exec_id,
                teardown=self._teardown,
                kwargs=comp_kwargs,
                result=self._result,
                errors=self._setup_errors,
            )
            if not self.is_bootstrap:
                self.block_forge_teardown()

        self.mark_as_executed()

    def teardown(self) -> None:
        logger.debug(
            f"Teardown task\n\t_exec_id: {self._exec_id}\n\tforge: {self.forge_full_path},\n\tscope: {self.forge_scope},\n\ttask: {self.test_key}\n\tteardown {self._teardown}"
        )
        try:
            teardown_start_time = time.time()
            if self._exec_id is not None and self._forge.teardown(
                self._exec_id
            ):
                logger.info(
                    f"Forge teardown has been executed successfully, time taken {time.time() - teardown_start_time} seconds:{self.summary}"
                )
        except Exception as e:
            traceback_info = traceback.format_exc()
            report = f"Forge teardown has failed to execute: {e}{self.summary}\n{traceback_info}"
            logger.error(report)
            self._teardown_errors.append(report)


TaskSetListType = List[Optional[List[FrameworkTask]]]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

UCC_MODINPUT_TEST_LOGGER_NAME = "splunk-add-on-ucc-modinput-test-functional"
UCC_MODINPUT_TEST_LOG_FILE_NAME = f"{UCC_MODINPUT_TEST_LOGGER_NAME}.log"

logFormatter = logging.Formatter(
    "%(asctime)s,%(msecs)03d %(levelname)s %(name)s pid=%(process)d tid=%(thread)d file=%(filename)s func=%(funcName)s line=%(lineno)d %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

fileHandler = logging.FileHandler(UCC_MODINPUT_TEST_LOG_FILE_NAME)
fileHandler.setFormatter(logFormatter)

logger = logging.getLogger(UCC_MODINPUT_TEST_LOGGER_NAME)
logger.addHandler(fileHandler)
logger.setLevel(logging.DEBUG)



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time
import random
import string
from enum import Enum


class IdentifierType(Enum):
    NUMERIC = 1
    ALPHA = 2
    ALPHANUM = 3
    HEX = 4


def _convert_to_anybase(number: int, base: str) -> str:
    size = len(base)
    if number == 0:
        return base[0]

    code = []
    while number > 0:
        number, r = divmod(number, size)
        code.append(base[r])

    return "".join(code)


def create_identifier(
    *,
    id_type: IdentifierType = IdentifierType.ALPHANUM,
    in_uppercase: bool = False
) -> str:
    time_based = int(time.time() * 10**5) % 10**10
    identifier = time_based * 10**3 + random.randint(0, 10**3)
    if id_type == IdentifierType.NUMERIC:
        return str(identifier)

    if id_type == IdentifierType.HEX:
        hex_identifier = hex(identifier)
        return (
            hex_identifier.upper() if in_uppercase else hex_identifier.lower()
        )

    if id_type == IdentifierType.ALPHA:
        base = string.ascii_lowercase
    elif id_type == IdentifierType.ALPHANUM:
        base = string.digits + string.ascii_lowercase

    other = _convert_to_anybase(identifier, base)
    return other.upper() if in_uppercase else other.lower()



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from argparse import ArgumentTypeError
from splunk_add_on_ucc_modinput_test.functional.constants import (
    ForgeProbe,
    TasksWait,
    Executor,
)
from pytest import Parser

from typing import Callable, Union


def int_range(min: int, max: int) -> Callable[[Union[int, str]], int]:
    def int_range_validator(value: Union[int, str]) -> int:
        try:
            value = int(value)
            assert min <= value <= max
        except Exception as e:
            err = f'Invalid value "{value}". Expected integer value in \
                inclusive range [{min}, {max}]'
            raise ArgumentTypeError(err) from e
        return value

    return int_range_validator


def pytest_addoption(parser: Parser) -> None:
    splunk_group = parser.getgroup(
        "addonfactory-ucc-test/functional - Options for unified functional \
            test framework for Splunk technical add-ons"
    )

    splunk_group.addoption(
        "--sequential-execution",
        dest="sequential_execution",
        action="store_true",
        default=False,
        help="Use no threading (for debugging).",
    )

    splunk_group.addoption(
        "--do-not-fail-with-teardown",
        dest="do_not_fail_with_teardown",
        action="store_true",
        default=False,
        help="Do not fail test if test's teardown fails. By default a test \
            will fail if any of its forges teardowns fail, even if the test \
                itself passed.",
    )

    splunk_group.addoption(
        "--do-not-delete-at-teardown",
        dest="do_not_delete_at_teardown",
        action="store_true",
        default=False,
        help="Do not delete created resoueces at teardown. This flag is for \
            debug purposes. Based on this flag developers can add alternative \
                code to forges, that, for example, would disable imputs \
                    instead of deletng them in order to study inputs after \
                        tests execution.",
    )

    allowed_range = [
        Executor.MIN_THREAD_NUMBER.value,
        Executor.MAX_THREAD_NUMBER.value,
    ]
    default = Executor.DEFAULT_THREAD_NUMBER.value
    splunk_group.addoption(
        "--number-of-threads",
        dest="number_of_threads",
        type=int_range(*allowed_range),
        default=10,
        help=f"Number of threads to use to execute forges. Allowed range: \
            {allowed_range}. Default value: {default}.",
    )

    allowed_range = [
        ForgeProbe.MIN_INTERVAL.value,
        ForgeProbe.MAX_INTERVAL.value,
    ]
    default = ForgeProbe.DEFAULT_INTERVAL.value
    splunk_group.addoption(
        "--probe-invoke-interval",
        dest="probe_invoke_interval",
        type=int_range(*allowed_range),
        default=default,
        help=f"Interval in seconds used to repeat invocation of yes/no type of \
            probe. Allowed range: {allowed_range}. Default value: {default}.",
    )

    allowed_range = [
        ForgeProbe.MIN_WAIT_TIME.value,
        ForgeProbe.MAX_WAIT_TIME.value,
    ]
    default = ForgeProbe.DEFAILT_WAIT_TIME.value
    splunk_group.addoption(
        "--probe-wait-timeout",
        dest="probe_wait_timeout",
        type=int_range(*allowed_range),
        default=default,
        help=f"Maximum time in seconds given to single probe to turn positive. \
            Allowed range: {allowed_range}. Default value: {default}.",
    )

    allowed_range = [
        TasksWait.MIN_BOOTSTRAP_TIMEOUT.value,
        TasksWait.MAX_BOOTSTRAP_TIMEOUT.value,
    ]
    default = TasksWait.DEFAULT_BOOTSTRAP_TIMEOUT.value
    splunk_group.addoption(
        "--bootstrap-wait-timeout",
        dest="bootstrap_wait_timeout",
        type=int_range(*allowed_range),
        default=default,
        help=f"Maximum time in seconds given to all bootstrap tasks to finish. \
            Allowed range: {allowed_range}. Default value: {default}.",
    )

    allowed_range = [
        TasksWait.MIN_ATTACHED_TIMEOUT.value,
        TasksWait.MAX_ATTACHED_TIMEOUT.value,
    ]
    default = TasksWait.DEFAULT_ATTACHED_TIMEOUT.value
    splunk_group.addoption(
        "--attached-tasks-wait-timeout",
        dest="attached_tasks_wait_timeout",
        type=int_range(*allowed_range),
        default=default,
        help=f"Maximum time in seconds given to finish all tasks attached to \
            a test. Allowed range: {allowed_range}. Default value: {default}.",
    )

    allowed_range = [
        TasksWait.MIN_CHECK_FREQUENCY.value,
        TasksWait.MAX_CHECK_FREQUENCY.value,
    ]
    default = TasksWait.DEFAULT_CHECK_FREQUENCY.value
    splunk_group.addoption(
        "--completion-check-frequency",
        dest="completion_check_frequency",
        type=int_range(*allowed_range),
        default=default,
        help=f"Frequency to check that bootstrap or attached tasks bundle has \
            finished to execute. Allowed range: {allowed_range}. \
                Default value: {default}.",
    )



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, List
from types import ModuleType
from splunk_add_on_ucc_modinput_test.common import utils
from splunk_add_on_ucc_modinput_test.common.splunk_instance import (
    Configuration as SplunkConfiguration,
)


class InputConfigurationBase:
    def __init__(self, *, name_prefix: str, interval: int = 60):
        self._name = f"{name_prefix}{utils.Common().sufix}"
        self._interval = interval
        self._index = SplunkConfiguration().dedicated_index.name

    @property
    def name(self) -> str:
        return self._name

    @property
    def interval(self) -> int:
        return self._interval

    @property
    def index(self) -> str:
        return self._index


class ConfigurationBase:
    def __init__(
        self,
        *,
        swagger_client: ModuleType,
        splunk_configuration: SplunkConfiguration,
    ):
        self.splunk_configuration = splunk_configuration

        configuration = swagger_client.Configuration()
        configuration.host = configuration.host.replace(
            "{domain}", self.splunk_configuration.host
        )
        configuration.host = configuration.host.replace(
            "{port}", self.splunk_configuration.port
        )

        configuration.verify_ssl = False
        configuration.username = self.splunk_configuration.username
        configuration.password = self.splunk_configuration.password

        self._api_instance = swagger_client.DefaultApi(
            swagger_client.ApiClient(configuration)
        )
        self._inputs: Dict[str, InputConfigurationBase] = {}

        def api_instance(self) -> swagger_client.api.default_api.DefaultApi:  # type: ignore # noqa: E501
            return self._api_instance

        ConfigurationBase.api_instance = property(api_instance)  # type: ignore

    @property
    def dedicated_index_name(self) -> str:
        return self.splunk_configuration.dedicated_index.name

    def add_input_configuration(
        self, input_configuration: InputConfigurationBase
    ) -> None:
        self._inputs[input_configuration.name] = input_configuration

    def get_all_inputs(self) -> List[InputConfigurationBase]:
        return [i for i in self._inputs.values()]

    def get_input_configuration(
        self, name_prefix: str
    ) -> InputConfigurationBase:
        return self._inputs[f"{name_prefix}{utils.Common().sufix}"]



#
# Copyright 2025 Splunk Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List
import pytest
import traceback
from splunk_add_on_ucc_modinput_test.functional import logger
from splunk_add_on_ucc_modinput_test.functional.exceptions import (
    SplTaFwkBaseException,
)
from splunk_add_on_ucc_modinput_test.functional.manager import (
    dependency_manager,
)
from splunk_add_on_ucc_modinput_test.functional.pytest_plugin.utils import (
    _adjust_test_order,
    _debug_log_test_order,
    _log_test_order,
    _collect_skipped_tests,
    _collect_parametrized_tests,
    _extract_parametrized_data,
    _map_forged_tests_to_pytest_items,
    _check_session_terminal_output,
)
from pytest import Session, Config, Item
from typing import Sequence


@pytest.hookimpl
def pytest_deselected(items: Sequence[Item]) -> None:
    logger.debug(f"Processing deselected items: {items}")
    if not items:
        return

    for item in items:
        found_tests_keys = (
            dependency_manager.tests.lookup_by_original_function(item._obj)
        )
        for test_key in found_tests_keys:
            test = dependency_manager.unregister_test(test_key)
            msg = "Test deselection:"
            msg += f'\n\tdeselected: {"Yes" if test else "No"}'
            msg += f"\n\tlookup key: {test_key}"
            msg += f'\n\tpath: {test.full_path if test else "not found"}'
            msg += f'\n\toriginal path: {test.original_full_path if test else "not found"}'
            logger.info(msg)


@pytest.hookimpl
def pytest_collection_modifyitems(
    session: Session, config: Config, items: List[Item]
) -> None:
    dependency_manager.link_pytest_config(config)
    if dependency_manager.collectonly:
        return

    logger.debug(f"Looking for forged tests in: {items}")
    tests2items = _map_forged_tests_to_pytest_items(items)
    if not tests2items:
        logger.debug("No forged tests found, exiting")
        return

    pytest_test_set_keys = set(tests2items.keys())
    dependency_manager.synch_tests_with_pytest_list(pytest_test_set_keys)

    parametrized_tests = _collect_parametrized_tests(items)
    dependency_manager.expand_parametrized_tests(parametrized_tests)
    dependency_manager.dump_tests()

    logger.debug("parametrized_tests: {parametrized_tests}")
    for test_key, param_tests in parametrized_tests.items():
        logger.debug(f"test: {test_key}")
        for test_name, test_kwargs in param_tests:
            logger.debug(
                f"\ttest name: {test_name}, test kwargs: {test_kwargs}"
            )

    skipped_tests = _collect_skipped_tests(items)
    skipped_tests_keys = [test.key for test, _ in skipped_tests]
    dependency_manager.remove_skipped_tests(skipped_tests_keys)

    items[:] = _adjust_test_order(items)

    _debug_log_test_order(items)
    _log_test_order(items)


@pytest.hookimpl
def pytest_collection_finish(session: Session) -> None:
    if dependency_manager.collectonly:
        return
    _check_session_terminal_output(session)

    dependency_manager.start_bootstrap_execution()


@pytest.hookimpl
def pytest_runtest_setup(item: Item) -> None:
    pytest_funcname, _ = _extract_parametrized_data(item)
    test = dependency_manager.find_test(item._obj, pytest_funcname)
    if not test:
        return

    logger.info(
        f"Executing pytest runtest setup step for forged test : {test}"
    )

    try:
        dependency_manager.wait_for_test_bootstrap(test)
        dependency_manager.execute_test_inplace_forges(test)
    except SplTaFwkBaseException as e:
        logger.error(f"Error during test setup: {e}\n{traceback.format_exc()}")
        pytest.fail(str(e))

    global_builtin_args = dependency_manager.get_global_builtin_args(test.key)
    custom_args = test.collect_required_kwargs(global_builtin_args)
    item.funcargs.update(custom_args)
    logger.info(
        f"pytest_runtest_setup: {item} ==>> {test},\n\tartifacts: {test.artifacts},\n\t_fixtureinfo: {item._fixtureinfo.argnames},\n\tfuncargs={item.funcargs},\n\tvars: {vars(item)}"
    )


@pytest.hookimpl
def pytest_runtest_call(item: Item) -> None:
    pytest_funcname, _ = _extract_parametrized_data(item)
    test = dependency_manager.find_test(item._obj, pytest_funcname)
    if not test:
        return

    logger.info(f"Executing pytest runtest call step for forged test : {test}")


@pytest.hookimpl
def pytest_runtest_teardown(item: Item) -> None:
    pytest_funcname, _ = _extract_parametrized_data(item)
    test = dependency_manager.find_test(item._obj, pytest_funcname)
    if not test:
        return

    logger.info(
        f"Executing pytest runtest teardown step for forged test : {test}"
    )
    dependency_manager.teardown_test(test)

    if dependency_manager.check_all_tests_executed():
        dependency_manager.shutdown()

    for task, error in dependency_manager.test_error_report(test):
        item.add_report_section("call", "error", str(error))

    if not dependency_manager.do_not_fail_with_teardown:
        msg = ""
        for task, error in dependency_manager.test_teardown_error_report(test):
            msg += (
                f"\n\tforge: {task.forge_full_path}, scope: {task.forge_scope}"
            )

        if msg:
            pytest.fail(f"teardown failed:\n\ttest: {test.key}{msg}")
